/*-------------------------------------------------------------------
Name......... audiohsfdrv.c
Author....... Ed Langley
Date......... 28/07/2005
Description.. Simplified driver for the Heber Sound Format (pre-
............. processed ULAW files) Installable from device list in
............. devices.c of project.
-------------------------------------------------------------------*/

/*-------------------------------------------------------------------
DECLARATION FORMATS
Type defs........... UPPER CASE
#defines............ UPPER CASE
Global Functions.... GlobalFunction();
Static Functions.... staticfunction();
Global Variables.... GlobalVariable;
Static Variables.... staticvariable;
Local Variables..... localvariable;
Structure Elements.. Structure.element;
-------------------------------------------------------------------*/

/****	Include Files */

#include	"hardware.h"
#include	"interface.h"
#include	"fpga.h"
#include	"fpga.p"
#include	"audiohsfdrv.sp"

/****	Externals */
/* 10ms interrupt external */
extern	void (*Int10msLst[])();
/* DMA interrupt external */
extern	void (*IntDMALst[])();

/* audio config external */
extern	const	AUDIOHSFCFG	AudioHSFCfg[];

/****	Local Definitions */
/* add two words together, don't divide */
#define SIGNEDWORDADD(x,y)  (WORD)((LONG)((LONG)(x)+(LONG)(y)))

/****	Global Variables */
UBYTE	HSFDriverInstalled = 0;

/****	Local Variables */
static	AUDIOHSFVARS AudioHSFVars={0};

/* one channel info structure for each channel */
static HSFCHANNELINFO HSFchannel[NUM_CHANNELS];

/* main driver state variable */
static volatile UBYTE hsfstate;

/* flag for any channels active */
static volatile UBYTE hsfchannelsactive;

static WORD dma_buffer[2][DMA_BUF_SIZE];

/* Output volume register */
static BYTE sfxvolregister=0;

/* mutex flags */
static volatile ULONG In10ms=FALSE;
static volatile ULONG ISR10msInterrupted=FALSE;

static UBYTE watchdog_hsfdriver;
static ULONG watchdog_hsfdriver_counter=0;
static UBYTE flag_HSFwatchdog_enable = FALSE;

	#define STATE_TRACE_BUF_SIZE	100
	/*uncomment this line if we want to debug the HSF driver
	in order to make checking trace buffer in debugger easier */
	//#define __DEBUG_HSF

	#ifdef __DEBUG_HSF
		STATEBUFFER statetracebuffer[STATE_TRACE_BUF_SIZE];
		UBYTE statetracebufix = 0;
	#endif



/****	Global Constants */
const	DEVICE	AudioHSFDevice=
{
	AUDIOHSFDEVICE,
	STDAUDIOHSF,
	-1,

	InstallAudioHSF,
	ShutdownAudioHSF,
	RestoreAudioHSF,

	&AudioHSFVars,
	0,
	0,
	AudioHSFCfg,
	0,
	0,
};

/****	Local Constants */
/*
rather than having a single hsfdonothing function,
we will have one function doing nothing for the dma and one for
the 10 ms, where we will generate a watch dog in case sound driver hangs
*/

static void hsfdmaerror();
static void hsf10mserror();
static void hsfdmadonothing();
static void hsf10msdonothing();

/* main driver state table */
struct hfsstates
{
	BYTE	next_state;
	UBYTE	dma_buf_tofill;
	UBYTE	audio_chan_tofill;
	void (*action)();
} static const HSFStateTable[64][2]={
/* states 						\events:AudioHSF10msISR								DMA2ISR */
/* statenum	dmabuf1 dmabuf0 hsfchan	|	n_state|cur_dma|cur_chan|action	| 			n_state|cur_dma|cur_chan|action */
/* 	0		EMPTY	EMPTY	0 		|*/{{5,		0,		0,		 loaddata},				{-1,	0,		0,		 hsfdmaerror},},
/* 	1		EMPTY	EMPTY	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	2		EMPTY	EMPTY	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	3		EMPTY	EMPTY	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},
/* 	4		EMPTY	FILLING 0		|*/{{-1,	0,		0,		 hsf10mserror},			{-1,	0,		0,		 hsfdmaerror},},
/* 	5		EMPTY	FILLING	1 		|*/{{6,		0,		1,		 loaddata},				{-1,	0,		1,		 hsfdmaerror},},
/* 	6		EMPTY	FILLING	2 		|*/{{7,		0,		2,		 loaddata},				{-1,	0,		2,		 hsfdmaerror},},
/* 	7		EMPTY	FILLING	3 		|*/{{12,	0,		3,		 loaddatabegindma},		{0,		0,		3,		 hsfdmadonothing},},
/* 	8		EMPTY	FULL	0 		|*/{{-1,	0,		0,		 hsf10mserror},			{-1,	0,		0,		 hsfdmaerror},},
/* 	9		EMPTY	FULL	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	10		EMPTY	FULL	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	11		EMPTY	FULL	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},
/* 	12		EMPTY	ACTIVE	0 		|*/{{29,	1,		0,		 loaddata},				{0,		1,		0,		 hsfdmadonothing},},
/* 	13		EMPTY	ACTIVE	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	14		EMPTY	ACTIVE	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	15		EMPTY	ACTIVE	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},
/* 	16		FILLING	EMPTY	0 		|*/{{-1,	0,		0,		 hsf10mserror},			{-1,	0,		0,		 hsfdmaerror},},
/* 	17		FILLING	EMPTY	1 		|*/{{18,	1,		1,		 loaddata},				{-1,	0,		1,		 hsfdmaerror},},
/* 	18		FILLING	EMPTY	2 		|*/{{19,	1,		2,		 loaddata},				{-1,	0,		2,		 hsfdmaerror},},
/* 	19		FILLING	EMPTY	3 		|*/{{48,	1,		3,		 loaddatabegindma},		{0,		0,		3,		 hsfdmadonothing},},
/* 	20		FILLING	FILLING 0 		|*/{{-1,	0,		0,		 hsf10mserror},			{-1,	0,		0,		 hsfdmaerror},},
/* 	21		FILLING	FILLING	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	22		FILLING	FILLING	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	23		FILLING	FILLING	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},
/* 	24		FILLING	FULL	0 		|*/{{-1,	0,		0,		 hsf10mserror},			{-1,	0,		0,		 hsfdmaerror},},
/* 	25		FILLING	FULL	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	26		FILLING	FULL	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	27		FILLING	FULL	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},
/* 	28		FILLING	ACTIVE	0 		|*/{{-1,	0,		0,		 hsf10mserror},			{-1,	0,		0,		 hsfdmaerror},},
/* 	29		FILLING	ACTIVE	1 		|*/{{30,	1,		1,		 loaddata},				{17,	1,		1,		 hsfdmadonothing},},
/* 	30		FILLING	ACTIVE	2 		|*/{{31,	1,		2,		 loaddata},				{18,	1,		2,		 hsfdmadonothing},},
/* 	31		FILLING	ACTIVE	3 		|*/{{44,	1,		3,		 loaddata},				{19,	1,		3,		 hsfdmadonothing},},
/* 	32		FULL	EMPTY	0 		|*/{{-1,	0,		0,		 hsf10mserror},			{-1,	0,		0,		 hsfdmaerror},},
/* 	33		FULL	EMPTY	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	34		FULL	EMPTY	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	35		FULL	EMPTY	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},
/* 	36		FULL	FILLING 0 		|*/{{-1,	0,		0,		 hsf10mserror},			{-1,	0,		0,		 hsfdmaerror},},
/* 	37		FULL	FILLING	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	38		FULL	FILLING	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	39		FULL	FILLING	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},
/* 	40		FULL	FULL	0 		|*/{{-1,	0,		0,		 hsf10mserror},			{-1,	0,		0,		 hsfdmaerror},},
/* 	41		FULL	FULL	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	42		FULL	FULL	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	43		FULL	FULL	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},
/* 	44		FULL	ACTIVE	0 		|*/{{44,	1,		0,		 hsf10msdonothing},		{48,	1,		0,		 switchdmabuf},},
/* 	45		FULL	ACTIVE	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	46		FULL	ACTIVE	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	47		FULL	ACTIVE	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},
/* 	48		ACTIVE	EMPTY	0 		|*/{{53,	0,		0,		 loaddata},				{0,		0,		0,		 hsfdmadonothing},},
/* 	49		ACTIVE	EMPTY	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	50		ACTIVE	EMPTY	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	51		ACTIVE	EMPTY	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},
/* 	52		ACTIVE	FILLING 0 		|*/{{-1,	0,		0,		 hsf10mserror},			{-1,	0,		0,		 hsfdmaerror},},
/* 	53		ACTIVE	FILLING	1 		|*/{{54,	0,		1,		 loaddata},				{5,		0,		1,		 hsfdmadonothing},},
/* 	54		ACTIVE	FILLING	2 		|*/{{55,	0,		2,		 loaddata},				{6,		0,		2,		 hsfdmadonothing},},
/* 	55		ACTIVE	FILLING	3 		|*/{{56,	0,		3,		 loaddata},				{7,		0,		3,		 hsfdmadonothing},},
/* 	56		ACTIVE	FULL	0 		|*/{{56,	0,		0,		 hsf10msdonothing},		{12,	0,		0,		 switchdmabuf},},
/* 	57		ACTIVE	FULL	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	58		ACTIVE	FULL	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	59		ACTIVE	FULL	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},
/* 	60		ACTIVE	ACTIVE	0 		|*/{{-1,	0,		0,		 hsf10mserror},			{-1,	0,		0,		 hsfdmaerror},},
/* 	61		ACTIVE	ACTIVE	1 		|*/{{-1,	0,		1,		 hsf10mserror},			{-1,	0,		1,		 hsfdmaerror},},
/* 	62		ACTIVE	ACTIVE	2 		|*/{{-1,	0,		2,		 hsf10mserror},			{-1,	0,		2,		 hsfdmaerror},},
/* 	63		ACTIVE	ACTIVE	3 		|*/{{-1,	0,		3,		 hsf10mserror},			{-1,	0,		3,		 hsfdmaerror},},

};


/*-------------------------------------------------------------------
Look-up table for U-law decompress

Generated using the following decompression algorithm:
Reference: sox.exe source code (sound processing utility)

#define QUANT_MASK	(0x0f)
#define BIAS		(0x84)
#define SEG_MASK	(0x70)
#define SEG_SHIFT	(4)
#define SIGN_BIT	(0x80)

short ulaw2linear(unsigned char u_val)
{
	int t;

	u_val = ~u_val;
	t = ((u_val & QUANT_MASK) <<3) + BIAS;
	t <<= ((unsigned)u_val & SEG_MASK) >> SEG_SHIFT;
	return( (u_val&SIGN_BIT) ? (BIAS-t) : (t-BIAS));
}

-------------------------------------------------------------------*/
static const WORD ulaw_lookup[ 256 ] =
{
    0x8284, 0x8684, 0x8a84, 0x8e84, 0x9284, 0x9684, 0x9a84, 0x9e84,
    0xa284, 0xa684, 0xaa84, 0xae84, 0xb284, 0xb684, 0xba84, 0xbe84,
    0xc184, 0xc384, 0xc584, 0xc784, 0xc984, 0xcb84, 0xcd84, 0xcf84,
    0xd184, 0xd384, 0xd584, 0xd784, 0xd984, 0xdb84, 0xdd84, 0xdf84,
    0xe104, 0xe204, 0xe304, 0xe404, 0xe504, 0xe604, 0xe704, 0xe804,
    0xe904, 0xea04, 0xeb04, 0xec04, 0xed04, 0xee04, 0xef04, 0xf004,
    0xf0c4, 0xf144, 0xf1c4, 0xf244, 0xf2c4, 0xf344, 0xf3c4, 0xf444,
    0xf4c4, 0xf544, 0xf5c4, 0xf644, 0xf6c4, 0xf744, 0xf7c4, 0xf844,
    0xf8a4, 0xf8e4, 0xf924, 0xf964, 0xf9a4, 0xf9e4, 0xfa24, 0xfa64,
    0xfaa4, 0xfae4, 0xfb24, 0xfb64, 0xfba4, 0xfbe4, 0xfc24, 0xfc64,
    0xfc94, 0xfcb4, 0xfcd4, 0xfcf4, 0xfd14, 0xfd34, 0xfd54, 0xfd74,
    0xfd94, 0xfdb4, 0xfdd4, 0xfdf4, 0xfe14, 0xfe34, 0xfe54, 0xfe74,
    0xfe8c, 0xfe9c, 0xfeac, 0xfebc, 0xfecc, 0xfedc, 0xfeec, 0xfefc,
    0xff0c, 0xff1c, 0xff2c, 0xff3c, 0xff4c, 0xff5c, 0xff6c, 0xff7c,
    0xff88, 0xff90, 0xff98, 0xffa0, 0xffa8, 0xffb0, 0xffb8, 0xffc0,
    0xffc8, 0xffd0, 0xffd8, 0xffe0, 0xffe8, 0xfff0, 0xfff8, 0x0000,
    0x7d7c, 0x797c, 0x757c, 0x717c, 0x6d7c, 0x697c, 0x657c, 0x617c,
    0x5d7c, 0x597c, 0x557c, 0x517c, 0x4d7c, 0x497c, 0x457c, 0x417c,
    0x3e7c, 0x3c7c, 0x3a7c, 0x387c, 0x367c, 0x347c, 0x327c, 0x307c,
    0x2e7c, 0x2c7c, 0x2a7c, 0x287c, 0x267c, 0x247c, 0x227c, 0x207c,
    0x1efc, 0x1dfc, 0x1cfc, 0x1bfc, 0x1afc, 0x19fc, 0x18fc, 0x17fc,
    0x16fc, 0x15fc, 0x14fc, 0x13fc, 0x12fc, 0x11fc, 0x10fc, 0x0ffc,
    0x0f3c, 0x0ebc, 0x0e3c, 0x0dbc, 0x0d3c, 0x0cbc, 0x0c3c, 0x0bbc,
    0x0b3c, 0x0abc, 0x0a3c, 0x09bc, 0x093c, 0x08bc, 0x083c, 0x07bc,
    0x075c, 0x071c, 0x06dc, 0x069c, 0x065c, 0x061c, 0x05dc, 0x059c,
    0x055c, 0x051c, 0x04dc, 0x049c, 0x045c, 0x041c, 0x03dc, 0x039c,
    0x036c, 0x034c, 0x032c, 0x030c, 0x02ec, 0x02cc, 0x02ac, 0x028c,
    0x026c, 0x024c, 0x022c, 0x020c, 0x01ec, 0x01cc, 0x01ac, 0x018c,
    0x0174, 0x0164, 0x0154, 0x0144, 0x0134, 0x0124, 0x0114, 0x0104,
    0x00f4, 0x00e4, 0x00d4, 0x00c4, 0x00b4, 0x00a4, 0x0094, 0x0084,
    0x0078, 0x0070, 0x0068, 0x0060, 0x0058, 0x0050, 0x0048, 0x0040,
    0x0038, 0x0030, 0x0028, 0x0020, 0x0018, 0x0010, 0x0008, 0x0000
};


/********************************************************************
*                                                                   *
* Driver setup Functions                                            *
*                                                                   *
********************************************************************/
/*-------------------------------------------------------------------
Name......... InstallAudioHSF()
Author....... Ed Langley
Date......... 28/07/2005
Description.. Install the HSF Audio device
Parameters... 1) Address of device
............. 2) Configuration data from the game level
............. 3) Base (unused - required by device manager)
Return....... TRUE on success or FALSE on failure
History...... 28/07/05 - Initial version - EL
-------------------------------------------------------------------*/
static BYTE		InstallAudioHSF(const DEVICE* dev,WORD cfg,WORD base)
{
	const AUDIOHSFCFG*	CfgPtr;
	AUDIOHSFVARS*		Vars;

	UBYTE channelindex;

	/* temporary bit mask which is shifted in the loop below, each
	   audio channel is then given a different active bit mask.
	   When channel is active these masks are then set in the global
	   active variable to enable faster checking:
	*/
	UBYTE tempactivemask;

	#ifdef __DEBUG_HSF
		UBYTE tracebufix;
	#endif

	CfgPtr = dev->config;
	CfgPtr = &CfgPtr[cfg];
	Vars = dev->vars;
	/* Set config pointer */
	Vars->config = CfgPtr;

	tempactivemask = 1;

	for(channelindex = 0; channelindex < NUM_CHANNELS; channelindex++)
	{
		/* initialise channel data */
		HSFchannel[channelindex].file_length = 0;
		HSFchannel[channelindex].currentpos = 0;
		HSFchannel[channelindex].looped = FALSE;
		HSFchannel[channelindex].dualmono = FALSE;
		HSFchannel[channelindex].status = CHANNEL_INACTIVE;
		HSFchannel[channelindex].active_mask = tempactivemask;
		tempactivemask <<= 1;
	}

	/* set the main driver state */
	hsfstate = 0; /* dma buf0/1 = empty, channel to fill = 0 */
	hsfchannelsactive = 0; /* no channels active */

	/* Enable SFX Audio Bit In FPGA */
	*Xcra |= Sfxenable;

	/* Set volume of both channels to zero */
	SetHSFChannelVolume( AUDIO_STEREO0, 0 );


	#ifdef __DEBUG_HSF
		for(tracebufix = 0; tracebufix < STATE_TRACE_BUF_SIZE; tracebufix++)
		{
			statetracebuffer[tracebufix].bufix = tracebufix;
		}
	#endif



	/* install interrupt routines */
	/* install hsf data loader into 10ms interupt list */
	if(InstallIntFnc(audiohsf10msisr, Int10msLst) == FALSE)
	{
		/* Interrupt list is full or function already installed */
		return FALSE;
	}

	/* install DMA isr into DMA interrupt list */
	if(InstallIntFnc(dmaisr, IntDMALst) == FALSE)
	{
		/* Interrupt list is full or function already installed */
		return FALSE;
	}

	watchdog_hsfdriver=0;

	HSFDriverInstalled = 1;
	return (TRUE);
}

/*-------------------------------------------------------------------
Name......... ShutdownAudioHSF()
Author....... Ed Langley
Date......... 28/07/2005
Description.. Put the HSF Audio device into shutdown
Parameters... 1) Address of device
Return....... None
History...... 28/07/05 - Initial version - EL
-------------------------------------------------------------------*/
static void		ShutdownAudioHSF(const DEVICE* dev)
{
}

/*-------------------------------------------------------------------
Name......... RestoreAudioHSF()
Author....... Ed Langley
Date......... 28/07/2005
Description.. Bring the HSF Audio device out of shutdown
Parameters... 1) Address of device
Return....... None
History...... 28/07/05 - Initial version - EL
-------------------------------------------------------------------*/
static void		RestoreAudioHSF(const DEVICE* dev)
{
}


/********************************************************************
*                                                                   *
* API Functions                                                     *
*                                                                   *
********************************************************************/
/*-------------------------------------------------------------------
Name......... OpenHSFile()
Author....... Ed Langley
Date......... 29/07/2005
Description.. Open a Heber Sound File.
Parameters... 1) Filename string
............. 2) Pointer to file descriptor
Return....... Success or failure code
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
UBYTE OpenHSFFile(const BYTE *filename, FILE_DESCRIPTOR *file)
{
	/* Open File */
	if ( FileOpen(AudioHSFVars.config->storagedevice, filename, file,
				   READ_ONLY) != 0 )
	{
		return(HSF_SUCCESS);
	}
	else
	{
		return(HSFERR_FILENOTFOUND);
	}
}

/*-------------------------------------------------------------------
Name......... PlayHSFile()
Author....... Ed Langley
Date......... 29/07/2005
Description.. Play a Heber Sound File.
Parameters... 1) channel to play on
............. 2) Pointer to file descriptor
Return....... Success or failure code
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
UBYTE PlayHSFFile(UBYTE channel, FILE_DESCRIPTOR *file)
{
	UBYTE tempchannel;

	/* remove the looped bit */
	tempchannel = channel & ~AUDIO_LOOPED;

	/* check if the sound channel is busy */
	if(HSFChannelBusy(tempchannel))
	{
		return HSFERR_CHANNELBUSY;
	}

	/* assign channels from definitions presented in the API */
	/* dualmono flag is for playing same mono data on both speakers */
	switch(tempchannel)
	{
		case AUDIO_LEFT0:
			tempchannel = HSF_CHAN_LEFT0;
			HSFchannel[tempchannel].dualmono = FALSE;
			break;
		case AUDIO_RIGHT0:
			tempchannel = HSF_CHAN_RIGHT0;
			HSFchannel[tempchannel].dualmono = FALSE;
			break;
		case AUDIO_LEFT1:
			tempchannel = HSF_CHAN_LEFT1;
			HSFchannel[tempchannel].dualmono = FALSE;
			break;
		case AUDIO_RIGHT1:
			tempchannel = HSF_CHAN_RIGHT1;
			HSFchannel[tempchannel].dualmono = FALSE;
			break;
		/* play the same mono data from a single file on both left
		   and right channels: */
		case AUDIO_STEREO0:
			tempchannel = HSF_CHAN_LEFT0;
			HSFchannel[tempchannel].dualmono = TRUE;
			break;
		case AUDIO_STEREO1:
			tempchannel = HSF_CHAN_LEFT1;
			HSFchannel[tempchannel].dualmono = TRUE;
			break;
		default:
			return HSFERR_INVALIDCHANNEL;
	}

	HSFchannel[tempchannel].looped = (channel & AUDIO_LOOPED) ? TRUE : FALSE;
	HSFchannel[tempchannel].file = file;
	HSFchannel[tempchannel].status = CHANNEL_STARTING;
	hsfchannelsactive |= HSFchannel[tempchannel].active_mask;
	return HSF_SUCCESS;

}

/*-------------------------------------------------------------------
Name......... PlayStereoHSFiles()
Author....... Ed Langley
Date......... 29/07/2005
Description.. Play left and right Heber Sound File pair in stereo.
Parameters... 1) channel to play on (must be STEREO0/1)
............. 2) Pointer to left channel file descriptor
............. 3) Pointer to right channel file descriptor
Return....... Success or failure code
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
UBYTE PlayStereoHSFFiles(UBYTE channel, FILE_DESCRIPTOR *leftfile,
									  FILE_DESCRIPTOR *rightfile)
{
	UBYTE tempchannel;
	UBYTE tempchannelleft;
	UBYTE tempchannelright;

	/* remove the looped bit */
	tempchannel = channel & ~AUDIO_LOOPED;

	/* check if the sound channel is busy */
	if(HSFChannelBusy(tempchannel))
	{
		return HSFERR_CHANNELBUSY;
	}

	switch(tempchannel)
	{
		case AUDIO_STEREO0:
			tempchannelleft = HSF_CHAN_LEFT0;
			tempchannelright = HSF_CHAN_RIGHT0;
			break;
		case AUDIO_STEREO1:
			tempchannelleft = HSF_CHAN_LEFT1;
			tempchannelright = HSF_CHAN_RIGHT1;
			break;
		default:
			return HSFERR_INVALIDCHANNEL;
	}

	/* dualmono flag is for playing same mono data on both speakers */
	HSFchannel[tempchannelleft].dualmono = FALSE;
	HSFchannel[tempchannelleft].looped = (channel & AUDIO_LOOPED) ? TRUE : FALSE;
	HSFchannel[tempchannelleft].file = leftfile;
	hsfchannelsactive |= HSFchannel[tempchannelleft].active_mask;

	HSFchannel[tempchannelright].dualmono = FALSE;
	HSFchannel[tempchannelright].looped = (channel & AUDIO_LOOPED) ? TRUE : FALSE;
	HSFchannel[tempchannelright].file = rightfile;
	hsfchannelsactive |= HSFchannel[tempchannelright].active_mask;

	/* block here until current channel to be filled = 0 to ensure
		synchronous playback */
	while(HSFStateTable[hsfstate][HSF_10MS_ISR].audio_chan_tofill);

	/* now set the status of the channels to have them start
		playback together */
	HSFchannel[tempchannelleft].status = CHANNEL_STARTING;
	HSFchannel[tempchannelright].status = CHANNEL_STARTING;

	return HSF_SUCCESS;
}

/*-------------------------------------------------------------------
Name......... StopHSFChannel()
Author....... Ed Langley
Date......... 29/07/2005
Description.. Stop playback of a Heber Sound File.
Parameters... 1) channel to stop
Return....... Success or failure code
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
UBYTE StopHSFChannel(UBYTE channel)
{
	UBYTE tempchannel;

	/* remove looped bit if it has been left in (shouldn't have) */
	tempchannel = channel & ~AUDIO_LOOPED;

	switch(tempchannel)
	{
		case AUDIO_LEFT0:
			tempchannel = HSF_CHAN_LEFT0;
			HSFchannel[tempchannel].status = CHANNEL_INACTIVE;
			break;
		case AUDIO_RIGHT0:
			tempchannel = HSF_CHAN_RIGHT0;
			HSFchannel[tempchannel].status = CHANNEL_INACTIVE;
			break;
		case AUDIO_LEFT1:
			tempchannel = HSF_CHAN_LEFT1;
			HSFchannel[tempchannel].status = CHANNEL_INACTIVE;
			break;
		case AUDIO_RIGHT1:
			tempchannel = HSF_CHAN_RIGHT1;
			HSFchannel[tempchannel].status = CHANNEL_INACTIVE;
			break;
		case AUDIO_STEREO0:
			StopHSFChannel(AUDIO_LEFT0);
			StopHSFChannel(AUDIO_RIGHT0);
			return 0; /* don't do file resetting at end of function
						 (tempchannel not valid) */
		case AUDIO_STEREO1:
			StopHSFChannel(AUDIO_LEFT1);
			StopHSFChannel(AUDIO_RIGHT1);
			return 0;
		default:
			return HSFERR_INVALIDCHANNEL;
	}

	/* stop the channel and reset the file */
	hsfchannelsactive &= ~HSFchannel[tempchannel].active_mask;
	FileRewind(HSFchannel[tempchannel].file);
	HSFchannel[tempchannel].currentpos = 0;


	return HSF_SUCCESS;
}

/*-------------------------------------------------------------------
Name......... HSFChannelBusy()
Author....... Ed Langley
Date......... 29/07/2005
Description.. Check status of an HSF channel.
Parameters... 1) channel to check
Return....... 1 if channel is busy, 0 if free
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
UBYTE HSFChannelBusy(UBYTE channel)
{
	UBYTE tempchannel;

	/* remove looped bit if it has been left in (shouldn't have) */
	tempchannel = channel & ~AUDIO_LOOPED;

	switch(tempchannel)
	{
		case AUDIO_LEFT0:
			if(HSFchannel[HSF_CHAN_LEFT0].status != CHANNEL_INACTIVE)
			{
				return HSFERR_CHANNELBUSY;
			}
			break;
		case AUDIO_RIGHT0:
			if( (HSFchannel[HSF_CHAN_RIGHT0].status != CHANNEL_INACTIVE) ||
				( (HSFchannel[HSF_CHAN_LEFT0].status != CHANNEL_INACTIVE) &&
				  (HSFchannel[HSF_CHAN_LEFT0].dualmono == TRUE) ) )
			{
				return HSFERR_CHANNELBUSY;
			}
			break;
		case AUDIO_LEFT1:
			if(HSFchannel[HSF_CHAN_LEFT1].status != CHANNEL_INACTIVE)
			{
				return HSFERR_CHANNELBUSY;
			}
			break;
		case AUDIO_RIGHT1:
			if( (HSFchannel[HSF_CHAN_RIGHT1].status != CHANNEL_INACTIVE) ||
				( (HSFchannel[HSF_CHAN_LEFT1].status != CHANNEL_INACTIVE) &&
				  (HSFchannel[HSF_CHAN_LEFT1].dualmono == TRUE) ) )
			{
				return HSFERR_CHANNELBUSY;
			}
			break;
		case AUDIO_STEREO0:
			if( (HSFchannel[HSF_CHAN_LEFT0].status != CHANNEL_INACTIVE) ||
				(HSFchannel[HSF_CHAN_RIGHT0].status != CHANNEL_INACTIVE) )
			{
				return HSFERR_CHANNELBUSY;
			}
			break;
		case AUDIO_STEREO1:
			if( (HSFchannel[HSF_CHAN_LEFT1].status != CHANNEL_INACTIVE) ||
				(HSFchannel[HSF_CHAN_RIGHT1].status != CHANNEL_INACTIVE) )
			{
				return HSFERR_CHANNELBUSY;
			}
			break;
		default:
			return HSFERR_INVALIDCHANNEL;
			break;
	}

	return HSF_SUCCESS;
}

/*-------------------------------------------------------------------
Name......... SetHSFChannelVolume()
Author....... Ed Langley
Date......... 29/07/2005
Description.. Set the volume of an HSF channel.
Parameters... 1) channel to set
............. 2) volume to set the channel to.
Return....... Success or failure code
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
UBYTE SetHSFChannelVolume(UBYTE channel, UBYTE volume)
{
	BYTE reqVolume;
	BYTE volLeft;
	BYTE volRight;

	/* Change 256 Levels to 16 */
	reqVolume = ( volume >> 4 ) & LOWER_NIBBLE_MASK;

	/* Store Current Volumes */
	volLeft = sfxvolregister & UPPER_NIBBLE_MASK;
	volRight = sfxvolregister & LOWER_NIBBLE_MASK;

	switch (channel)
	{
		case AUDIO_LEFT0:
		case AUDIO_LEFT1:
			volLeft = ( (reqVolume & LOWER_NIBBLE_MASK) * 16 );
			break;

		case AUDIO_RIGHT0:
		case AUDIO_RIGHT1:
			volRight = reqVolume & LOWER_NIBBLE_MASK;
			break;

		case AUDIO_STEREO0:
		case AUDIO_STEREO1:
			volLeft = ( (reqVolume & LOWER_NIBBLE_MASK) * 16 );
			volRight = reqVolume & LOWER_NIBBLE_MASK;
			break;

		default:
			return HSFERR_INVALIDCHANNEL;
			break;
	}

	/* Update Volume Register with new value */
	sfxvolregister = ( volLeft | volRight );
	*Xsfxvol = sfxvolregister;

	return HSF_SUCCESS;
}

/********************************************************************
*                                                                   *
* Driver Functions                                                  *
*                                                                   *
********************************************************************/
/*-------------------------------------------------------------------
Name......... loaddata()
Author....... Ed Langley
Date......... 29/07/2005
Description.. Load SECTORS_TO_LOAD number of sectors into the file
............. buffer of the current audio channel, then decompress
............. into the non-active dma buffer.
Parameters... None
Return....... None
History...... 29/07/05 - Initial version - EL
............. 05/09/05 - Zero DMA added on first loaddata call - AD
-------------------------------------------------------------------*/
static void loaddata()
{

	UBYTE			tempchannel;
	UBYTE			*filebufptr;
	UBYTE			*filebufendptr;
	WORD			*dmabufptr;
	ULONG			*lengthptr;
	WORD			mixsample;
	UBYTE			sectorstoload;

	/*
	The full loading of the four possible channels is happening
	on four 10 ms interrupts.
	The dma buffer to fill will be picked on first 10ms interrupt.
	Data will be inserted within the same DMA buffer for each channel on
	each 10ms interrupt.

	The first channel to be filled after a DMA has  occured is HSF_CHAN_LEFT0
	This is the only time the dma buffer can be cleared.
	*/

	if(HSFStateTable[hsfstate][HSF_10MS_ISR].audio_chan_tofill == HSF_CHAN_LEFT0)
	{
		if(HSFStateTable[hsfstate][HSF_10MS_ISR].dma_buf_tofill)
		{
			zeroDMAbuf(dma_buffer[1]);
		}
		else
		{
			zeroDMAbuf(dma_buffer[0]);
		}
	}

	/* check if the channel is active */
	if(HSFchannel[HSFStateTable[hsfstate][HSF_10MS_ISR].audio_chan_tofill].status == CHANNEL_INACTIVE)
	{
		return;
	}

	/* reset watchdog */
	watchdog_hsfdriver=0;


	/* take local copies of certain variables. A change in state by
	   the DMA interrupt will not change the value of these: */
	/* audio channel to fill */
	tempchannel = HSFStateTable[hsfstate][HSF_10MS_ISR].audio_chan_tofill;
	/* pointer to the data area to read the file sectors to */
	filebufptr = HSFchannel[tempchannel].fileBuffer.bytes;
	/* pointer to the DMA buffer to fill */
	dmabufptr = &dma_buffer[HSFStateTable[hsfstate][HSF_10MS_ISR].dma_buf_tofill][0];

	/* Determine how much data needs to be loaded, must check for channel starting first
	   in order to get a valid file length value: */
	if(HSFchannel[tempchannel].status == CHANNEL_STARTING)
	{
		/* load required amount of sectors to fill 45ms of audio */
		HSFchannel[tempchannel].currentpos += SECTORS_TO_LOAD;
		FileGetSectors(AudioHSFVars.config->storagedevice,
			   HSFchannel[tempchannel].file,
			   (UWORD*)(HSFchannel[tempchannel].fileBuffer.words),
			   SECTORS_TO_LOAD);

		/* Pointer to the end of the compressed data area */
		filebufendptr = HSFchannel[tempchannel].fileBuffer.bytes + FILE_BUF_SIZE;

		/* This is the start of the file, so deal with the length field */
		lengthptr = (ULONG*)(HSFchannel[tempchannel].fileBuffer.longs);
		HSFchannel[tempchannel].file_length = *lengthptr;
		filebufptr += FILE_LENGTH_FIELD_SIZE_BYTES;
		dmabufptr += FILE_LENGTH_FIELD_SIZE_WORDS;
		HSFchannel[tempchannel].status = CHANNEL_ACTIVE;

	}
	else if((HSFchannel[tempchannel].currentpos + SECTORS_TO_LOAD) > HSFchannel[tempchannel].file_length)
	{
		/* The end of the file is within less than SECTORS_TO_LOAD, in which case
		   only read the required number of sectors from the file to avoid wasting
		   CPU time, and in turn only decompress the required amount of data from
		   the file buffer to the DMA buffer, otherwise a "click" may be heard (Q1889).
	*/

		/* Only load required amount of sectors to read the rest of the file */
		sectorstoload = (HSFchannel[tempchannel].file_length - HSFchannel[tempchannel].currentpos);
		HSFchannel[tempchannel].currentpos += sectorstoload;
		FileGetSectors(AudioHSFVars.config->storagedevice,
			   HSFchannel[tempchannel].file,
			   (UWORD*)(HSFchannel[tempchannel].fileBuffer.words),
			   sectorstoload);

		/* Pointer to the end of the compressed data area */
		filebufendptr = HSFchannel[tempchannel].fileBuffer.bytes + (sectorstoload * BYTES_PER_SECTOR);
	}
	else
	{
		/* normal case, load required amount of sectors to fill 45ms of audio */
		HSFchannel[tempchannel].currentpos += SECTORS_TO_LOAD;
		FileGetSectors(AudioHSFVars.config->storagedevice,
			   HSFchannel[tempchannel].file,
			   (UWORD*)(HSFchannel[tempchannel].fileBuffer.words),
			   SECTORS_TO_LOAD);

		/* Pointer to the end of the compressed data area */
		filebufendptr = HSFchannel[tempchannel].fileBuffer.bytes + FILE_BUF_SIZE;
	}


	/*  perform ULAW lookup, copy the data into the nonactive dma buffer,
		and mix if neccessary. Right channel data goes first in DMA buffer
		on Pluto 6. All samples are halved (shifted down) before loading,
		then mixing is just an add. This avoids volume dips. */
	switch(tempchannel)
	{
		case HSF_CHAN_LEFT0:
			/* no need to mix any data as this is the first left channel to be loaded */
			if(HSFchannel[tempchannel].dualmono)
			{
				while(filebufptr < filebufendptr)
				{
					/* load the same data into both channels */
					*dmabufptr++ = (ulaw_lookup[*filebufptr] >> 2);
					*dmabufptr++ = (ulaw_lookup[*filebufptr++] >> 2);
				}
			}
			else
			{
				while(filebufptr < filebufendptr)
				{
					/* load the data into the left channel only */
					dmabufptr++;
					*dmabufptr++ = (ulaw_lookup[*filebufptr++] >> 2);
				}
			}
		break;
		case HSF_CHAN_RIGHT0:
			/* no need to mix any data as this is the first right channel to be loaded */
			while(filebufptr < filebufendptr)
			{
				/* load the data into the right channel only */
				*dmabufptr++ = (ulaw_lookup[*filebufptr++] >> 2);
				dmabufptr++;
			}
		break;
		case HSF_CHAN_LEFT1:
			if(HSFchannel[tempchannel].dualmono)
			{
				/* load the same data into both channels */
				if(HSFchannel[HSF_CHAN_LEFT0].status)
				{
					if( (HSFchannel[HSF_CHAN_LEFT0].dualmono) || (HSFchannel[HSF_CHAN_RIGHT0].status) )
					{
						while(filebufptr < filebufendptr)
						{
							/* load and mix both channels */
							mixsample = SIGNEDWORDADD(*dmabufptr, (ulaw_lookup[*filebufptr] >> 2));
							*dmabufptr++ = mixsample;
							mixsample = SIGNEDWORDADD(*dmabufptr, (ulaw_lookup[*filebufptr++] >> 2));
							*dmabufptr++ = mixsample;
						}
					}
					else
					{
						while(filebufptr < filebufendptr)
						{
							/* load both channels, just mix the left channel */
							*dmabufptr++ = (ulaw_lookup[*filebufptr] >> 2);
							mixsample = SIGNEDWORDADD(*dmabufptr, (ulaw_lookup[*filebufptr++] >> 2));
							*dmabufptr++ = mixsample;
						}
					}
				}
				else if(HSFchannel[HSF_CHAN_RIGHT0].status)
				{
					while(filebufptr < filebufendptr)
					{
						/* load both channels, just mix the right channel */
						mixsample = SIGNEDWORDADD(*dmabufptr, (ulaw_lookup[*filebufptr] >> 2));
						*dmabufptr++ = mixsample;
						*dmabufptr++ = (ulaw_lookup[*filebufptr++] >> 2);
					}
				}
				else
				{
					while(filebufptr < filebufendptr)
					{
						/* load both channels, no mixing required */
						*dmabufptr++ = (ulaw_lookup[*filebufptr] >> 2);
						*dmabufptr++ = (ulaw_lookup[*filebufptr++] >> 2);
					}
				}
			}
			else if(HSFchannel[HSF_CHAN_LEFT0].status)
			{
				while(filebufptr < filebufendptr)
				{
					/* load left channel and mix */
					dmabufptr++;
					mixsample = SIGNEDWORDADD(*dmabufptr, (ulaw_lookup[*filebufptr++] >> 2));
					*dmabufptr++ = mixsample;
				}

			}
			else
			{
				while(filebufptr < filebufendptr)
				{
					/* load left channel, no mixing required */
					dmabufptr++;
					*dmabufptr++ = (ulaw_lookup[*filebufptr++] >> 2);
				}
			}
		break;
		case HSF_CHAN_RIGHT1:
			if( (HSFchannel[HSF_CHAN_RIGHT0].status) ||
				((HSFchannel[HSF_CHAN_LEFT0].status) && (HSFchannel[HSF_CHAN_LEFT0].dualmono)) )
			{
				while(filebufptr < filebufendptr)
				{
					/* load right channel and mix */
					mixsample = SIGNEDWORDADD(*dmabufptr, (ulaw_lookup[*filebufptr++] >> 2));
					*dmabufptr++ = mixsample;
					dmabufptr++;
				}
			}
			else
			{
				while(filebufptr < filebufendptr)
				{
					/* load right channel, no mixing required */
					*dmabufptr++ = (ulaw_lookup[*filebufptr++] >> 2);
					dmabufptr++;
				}
			}
		break;
		default:
			/* check the state table if arriving here */
		break;

	}

	/* check if the end of the file has been reached */
	if(HSFchannel[tempchannel].currentpos >= HSFchannel[tempchannel].file_length)
	{
		/* reset the file */
		FileRewind(HSFchannel[tempchannel].file);
		HSFchannel[tempchannel].currentpos = 0;

		if(HSFchannel[tempchannel].looped)
		{
			/* set the staus to deal with the length field
			   at the start of the file again next time */
			HSFchannel[tempchannel].status = CHANNEL_STARTING;
		}
		else
		{
			/* stop the channel and reset the file */
			hsfchannelsactive &= ~HSFchannel[tempchannel].active_mask;
			HSFchannel[tempchannel].status = CHANNEL_INACTIVE;
			FileRewind(HSFchannel[tempchannel].file);
			HSFchannel[tempchannel].currentpos = 0;
		}
	}
}

/*-------------------------------------------------------------------
Name......... loaddatabegindma()
Author....... Ed Langley
Date......... 29/07/2005
Description.. Load a number of sectors into the file buffer of the
............. current audio channel, then decompress into the
............. non-active dma buffer, then begin dma.
Parameters... None
Return....... None
History...... 29/07/05 - Initial version - EL
............. 05/12/05 - Zero DMA removed - AD
-------------------------------------------------------------------*/
static void loaddatabegindma()
{
	loaddata();


	/* check which dma buf has been filled and start dma on that buffer */
	if(HSFStateTable[hsfstate][HSF_10MS_ISR].dma_buf_tofill)
	{
		/* start DMAing buf 1 */
		StartDMA2((BYTE*)dma_buffer[1], Sfxr1, DMA_BUF_SIZE_BYTES);

	}
	else
	{
		/* start DMAing buf 0 */
		StartDMA2((BYTE*)dma_buffer[0], Sfxr1, DMA_BUF_SIZE_BYTES);
	}
}

/*-------------------------------------------------------------------
Name......... switchdmabuf()
Author....... Ed Langley
Date......... 29/07/2005
Description.. Switch the current active (and now empty) DMA buffer
............. for the newly filled one.
Parameters... None
Return....... None
History...... 29/07/05 - Initial version - EL
............. 05/12/05 - Zero DMA removed - AD
-------------------------------------------------------------------*/
static void switchdmabuf()
{
	/*
	we cannot zero buffer withing DMA interrupt . Highest interrupt
	level will delay the 1ms interrupt and cause flicker of the lamps
	it has to be done somewhere else.
	*/

	/* check which dma buf has been filled and start dma on that buffer */
	if(HSFStateTable[hsfstate][HSF_DMA_ISR].dma_buf_tofill)
	{
		/* start DMAing buf 1 */
		StartDMA2((BYTE*)dma_buffer[1], Sfxr1, DMA_BUF_SIZE_BYTES);
	}
	else
	{
		/* start DMAing buf 0 */
		StartDMA2((BYTE*)dma_buffer[0], Sfxr1, DMA_BUF_SIZE_BYTES);
	}

}

/*-------------------------------------------------------------------
Name......... zeroDMAbuf()
Author....... Ed Langley
Date......... 08/08/2005
Description.. The inactive dma buffer to avoid end of a sample
............. playing repeatedly.
Parameters... Pointer to DMA buffer to clear (0/1)
Return....... None
History...... 08/08/05 - Initial version - EL
-------------------------------------------------------------------*/
static void zeroDMAbuf(WORD *dma_buf_ptr)
{
	WORD *dmabufendptr;
	dmabufendptr = dma_buf_ptr + DMA_BUF_SIZE;
	while(dma_buf_ptr < dmabufendptr)
	{
		*dma_buf_ptr++ = 0;
	}
}

/*-------------------------------------------------------------------
Name......... hsferror()
Author....... Ed Langley
Date......... 29/07/2005
Description.. This is called from an invalid state in the table,
............. output an error and the state number.
Parameters... None
Return....... None
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
static void hsfdmaerror()
{

	asm("halt");
}




/*-------------------------------------------------------------------
Name......... hsferror()
Author....... Ed Langley
Date......... 29/07/2005
Description.. This is called from an invalid state in the table,
............. output an error and the state number.
Parameters... None
Return....... None
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
static void hsf10mserror()
{

	asm("halt");
}



/*-------------------------------------------------------------------
Name......... hsfdonothing()
Author....... Ed Langley
Date......... 29/07/2005
Description.. Function stub to satisfy call to action for valid
............. states in the table.
Parameters... None
Return....... None
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
static void hsfdmadonothing()
{
}

/*-------------------------------------------------------------------
Name......... hsfdonothing()
Author....... Ed Langley
Date......... 29/07/2005
Description.. Function stub to satisfy call to action for valid
............. states in the table.
Parameters... None
Return....... None
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
static void hsf10msdonothing()
{
	if(flag_HSFwatchdog_enable)
	{
		watchdog_hsfdriver++;

		/* 4 10ms interrupt have passed and the driver does nothing, it has hanged
		probably in state 44 or 56 where the dma buffer are suppose to switch */
		if(watchdog_hsfdriver>=4) /* the driver is hanging */
		{
			reinitialise_hsf_driver();
		}
	}
}

static void reinitialise_hsf_driver()
{
		UBYTE channelindex;

		/* temporary bit mask which is shifted in the loop below, each
		   audio channel is then given a different active bit mask.
		   When channel is active these masks are then set in the global
		   active variable to enable faster checking:
		*/
		UBYTE tempactivemask;

		tempactivemask = 1;

		for(channelindex = 0; channelindex < NUM_CHANNELS; channelindex++)
		{
			/* initialise channel data */
			HSFchannel[channelindex].file_length = 0;
			HSFchannel[channelindex].currentpos = 0;
			HSFchannel[channelindex].looped = FALSE;
			HSFchannel[channelindex].dualmono = FALSE;
			HSFchannel[channelindex].status = CHANNEL_INACTIVE;
			HSFchannel[channelindex].active_mask = tempactivemask;
			tempactivemask <<= 1;
		}

		/* set the main driver state */
		hsfstate = 0; /* dma buf0/1 = empty, channel to fill = 0 */
		hsfchannelsactive = 0; /* no channels active */

		/* watchdog */
		if(flag_HSFwatchdog_enable)
		{
			watchdog_hsfdriver=0;
			watchdog_hsfdriver_counter++;
		}

}




/********************************************************************
*                                                                   *
* Interrupt Functions                                               *
*                                                                   *
********************************************************************/
/*-------------------------------------------------------------------
Name......... AudioHSF10msISR()
Author....... Ed Langley
Date......... 29/07/2005
Description.. 10ms ISR to load data into the DMA buffer, one channel
............. loaded per interrupt.
Parameters... None
Return....... None
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
static void audiohsf10msisr()
{
	#ifdef __DEBUG_HSF
		statetracebuffer[statetracebufix].currentstate = hsfstate;
		statetracebuffer[statetracebufix].event = HSF_10MS_ISR;
		if(++statetracebufix >= STATE_TRACE_BUF_SIZE)
		{
			statetracebufix = 0;
		}
	#endif



	/* set the "semaphore" before processing the event and updating the state: */
	In10ms = TRUE;


	/* before starting to fill channels, check to see if any active */
	if( (HSFStateTable[hsfstate][HSF_10MS_ISR].audio_chan_tofill !=
		HSF_CHAN_LEFT0) || hsfchannelsactive)
	{

		HSFStateTable[hsfstate][HSF_10MS_ISR].action();
		if(HSFStateTable[hsfstate][HSF_10MS_ISR].next_state >= 0)
		{
			hsfstate = HSFStateTable[hsfstate][HSF_10MS_ISR].next_state;
		}
	}


	In10ms = FALSE;

	/* Now check if this function was interrupted by the DMA ISR: */
	if(ISR10msInterrupted == TRUE)
	{
		/*
		If so perform the code accomplished by the DMA ISR now instead:
		*/
		HSFStateTable[hsfstate][HSF_DMA_ISR].action();
		if(HSFStateTable[hsfstate][HSF_DMA_ISR].next_state >= 0)
		{
			hsfstate = HSFStateTable[hsfstate][HSF_DMA_ISR].next_state;
		}
		ISR10msInterrupted=FALSE;
	}
}

/*-------------------------------------------------------------------
Name......... DMA2ISR()
Author....... Ed Langley
Date......... 29/07/2005
Description.. DMA ISR occuring when the current active buffer has run
............. out, switch buffers if next one is ready.
Parameters... None
Return....... None
History...... 29/07/05 - Initial version - EL
-------------------------------------------------------------------*/
static void dmaisr()
{

	/* Stop DMA Channel */
	StopDMA2();

	#ifdef __DEBUG_HSF
		statetracebuffer[statetracebufix].currentstate = hsfstate;
		statetracebuffer[statetracebufix].event = HSF_DMA_ISR;
		if(++statetracebufix >= STATE_TRACE_BUF_SIZE)
		{
			statetracebufix = 0;
		}
	#endif



	/* check if this ISR is interrupting the lower priority 10ms ISR:	*/
	if(In10ms == TRUE)
	{
		/* In which case inform the 10ms interrupt and let it handle
		the event sequentially */
		ISR10msInterrupted = TRUE;
		return;
	}


	HSFStateTable[hsfstate][HSF_DMA_ISR].action();
	if(HSFStateTable[hsfstate][HSF_DMA_ISR].next_state >= 0)
	{
		hsfstate = HSFStateTable[hsfstate][HSF_DMA_ISR].next_state;
	}

}
/*-------------------------------------------------------------------*/
void EnableHSFWatchdog()
{
	flag_HSFwatchdog_enable=TRUE;
}

void DisableHSFWatchdog()
{
	flag_HSFwatchdog_enable=FALSE;
}
