/*-------------------------------------------------------------------
Name......... coin.c
Author....... Lee Weekes
Date......... 29/04/07
Description.. Interface to the peripheral cctalk coin driver
-------------------------------------------------------------------*/

/*-------------------------------------------------------------------
DECLARATION FORMATS
Type defs........... UPPER CASE
#defines............ UPPER CASE
Global Functions.... GlobalFunction();
Static Functions.... staticfunction();
Global Variables.... GlobalVariable;
Static Variables.... staticvariable;
Local Variables..... localvariable;
Structure Elements.. Structure.element;
-------------------------------------------------------------------*/
#include	"hardware.h"
#include	"peripherals.h"
#include	"interface.h"
#include	"coin.h"
#include	"coin.sp"

/*------------------------------------------------------------------
Name......... GetCoinCount()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Obtain the coin count for the specified coin type
Parameters... 1) The device object passed in
............. 2) The coin denomination to return
Return....... TRUE on success or FALSE on failure
History...... 26/03/07 - Initial version - LW
-------------------------------------------------------------------*/

UWORD GetCoinCount(const DEVICE* dev, UBYTE coin)
{
	const COINSERVICES* cser;

	cser=dev->services;

        return(cser->get_coin_count(dev, coin));
}

/*-------------------------------------------------------------------
Name......... GetCoinEvents()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Return the last event from the device if any.
Parameters... 1) The device object passed in
Return....... 0 if no events are pending, or the first event.
History...... 12/04/07 - Initial version - LW
--------------------------------------------------------------------*/
UBYTE GetCoinEvents(const DEVICE* dev)
{
	const COINSERVICES* cser;

	cser=dev->services;

	return(cser->get_coin_events(dev));
}

/*-------------------------------------------------------------------
Name......... GetCoinErrors()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Get the last error resported by the device
Parameters... 1) The device object passed in
Return....... The last error code returned or 0 if no error
History...... 26/03/07 - Initial version - LW
--------------------------------------------------------------------*/
UBYTE GetCoinErrors(const DEVICE* dev)
{
	const COINSERVICES* cser;

	cser=dev->services;

	return(cser->get_coin_errors(dev));
}

/*-------------------------------------------------------------------
Name......... ResetCoinDevice()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Perform a software reset of the device, note this will
............. cause the device to stop responding for 50ms.
Parameters... 1) The device object passed in
Return....... None
History...... 26/03/07 - Initial version - LW
--------------------------------------------------------------------*/
void ResetCoinDevice(const DEVICE* dev)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->reset(dev);
}

/*-------------------------------------------------------------------
Name......... SetPollingRate()
Author....... Lee Weekes
Date......... 27/03/2007
Description.. Set the polling rate for the coin acceptor although it
............. will be configured with the default polling.
Parameters... 1) The device object passed in.
............. 2) The polling rate in milliseconds.
Return....... None
History...... 27/03/07 - Initial version - LW
--------------------------------------------------------------------*/
void SetCoinPollingRate(const DEVICE* dev, UWORD pollingrate)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->set_polling_rate(dev, pollingrate);
}

/*-------------------------------------------------------------------
Name......... GetCoinStatus()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Obtain the status of the coin device
Parameters... 1) The device object passed in.
Return....... The status of the device
History...... 26/03/07 - Initial version - LW
--------------------------------------------------------------------*/
UWORD GetCoinStatus(const DEVICE* dev)
{
	const COINSERVICES* cser;

	cser=dev->services;

	return(cser->get_coin_status(dev));
}

/*-------------------------------------------------------------------
Name......... EnableCoinProcessing()
Author....... Lee Weekes
Date......... 26/04/2007
Description.. Enable the coin acceptor
Parameters... 1) The device object passed in.
Return....... TRUE on success or FALSE on failure
History...... 26/03/07 - Initial version - LW
--------------------------------------------------------------------*/
void EnableCoinProcessing(const DEVICE* dev)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->enable_coin_processing(dev);
}

/*-------------------------------------------------------------------
Name......... DisableCoinProcessing()
Author....... Lee Weekes
Date......... 26/04/2007
Description.. Enable the coin acceptor
Parameters... 1) The coin device object passed in.
Return....... TRUE on success or FALSE on failure
History...... 26/03/07 - Initial version - LW
--------------------------------------------------------------------*/
void DisableCoinProcessing(const DEVICE* dev)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->enable_coin_processing(dev);
}

/*-------------------------------------------------------------------
Name......... SubtractFromCoinCount()
Author....... Lee Weekes
Date......... 26/04/2007
Description.. Subtract one from the coin count for the selected coin
Parameters... 1) The device object passed in.
............. 2) The denomination to subtract.
Return....... None
History...... 26/03/07 - Initial version - LW
--------------------------------------------------------------------*/
void SubtractFromCoinCounts(const DEVICE* dev, UBYTE coin)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->subtract_from_coin_counts(dev, coin);
}

/*-------------------------------------------------------------------
Name......... AddToCoinCount()
Author....... Lee Weekes
Date......... 26/04/2007
Description.. Add one from the coin count for the selected coin
Parameters... 1) The device object passed in.
............. 2) The denomination to subtract from.
Return....... None
History...... 26/03/07 - Initial version - LW
--------------------------------------------------------------------*/
BYTE AddToCoinCounts(const DEVICE* dev, UBYTE coin)
{
	const COINSERVICES* cser;

	cser=dev->services;

	return cser->add_to_coin_counts(dev, coin);
}

/*-------------------------------------------------------------------
Name......... PerformCoinSelfTest()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Perform a self-test of the coin acceptor
Parameters... 1) The device object passed in.
Return....... TRUE on success or FALSE on failure
History...... 26/03/07 - Initial version - LW
--------------------------------------------------------------------*/
void PerformCoinSelfTest(const DEVICE* dev)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->selftest(dev);
}

/*-------------------------------------------------------------------
Name......... IsCoinEnabled()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Check to see if a specified coin is enabled
Parameters... 1) The device object passed in.
............. 2) The denomination to check.
Return....... TRUE if the coin is enabled, FALSE if not.
History...... 26/03/07 - Initial version - LW
-------------------------------------------------------------------	*/
UBYTE IsCoinEnabled(const DEVICE* dev, UBYTE coin)
{
	const COINSERVICES* cser;

	cser=dev->services;

	return(cser->is_coin_enabled(dev, coin));
}

/*-------------------------------------------------------------------
Name......... GetAllEnabledCoins()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Obtain the current inhibit status of the device
Parameters... 1) The device object passed in.
Return....... The current inhibit status of the device
History...... 26/03/07 - Initial version - LW
-------------------------------------------------------------------	*/
UWORD GetAllEnabledCoins(const DEVICE* dev)
{
	const COINSERVICES* cser;

	cser=dev->services;

	return(cser->get_all_enabled_coins(dev));
}

/*-------------------------------------------------------------------
Name......... GetSorterOverride()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Obtain the sorter override. This is a blocking function
............. as this should be called infrequently and we do not
............. want to have too many calls in the state machine due to
............. the coin acceptors having a high polling rate (200ms).
Parameters... 1) The device object passed in.
Return....... The current sorter override.
History...... 26/03/07 - Initial version - LW
-------------------------------------------------------------------	*/
UBYTE GetSorterOverride(const DEVICE* dev)
{
	const COINSERVICES* cser;

	cser=dev->services;

	return(cser->get_sorter_override(dev));
}

/*-------------------------------------------------------------------
Name......... SetSorterOverride()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Set the sorter override for the device
Parameters... 1) The device object passed in
............. 2) The new sorter override for the device
Return....... TRUE on success or FALSE on failure
History...... 26/03/07 - Initial version - LW
-------------------------------------------------------------------	*/
void SetSorterOverride(const DEVICE* dev, UBYTE override)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->set_sorter_override(dev, override);
}

/*-------------------------------------------------------------------
Name......... GetSorterPaths()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Obtain the sorter paths. This is a blocking function
............. as this should be called infrequently and we do not
............. want to have too many calls in the state machine due to
............. the coin acceptors having a high polling rate (200ms).
Parameters... 1) The device object passed in.
............. 2) The coin position to request the paths for.
............. 3) The paths to return
Return....... None
History...... 26/03/07 - Initial version - LW
............. 28/11/07 - Added required parameter paths
-------------------------------------------------------------------	*/
void GetSorterPaths(const DEVICE* dev, UBYTE coinposition, ULONG *paths)
{
	const COINSERVICES* cser;

	cser=dev->services;

        cser->get_sorter_paths(dev, coinposition, paths);
}

/*-------------------------------------------------------------------
Name......... SetSorterPaths()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Set the sorter paths for the specified device
Parameters... 1) The device object passed in
............. 2) The coin position to set the paths for
............. 3) The paths passed in
Return....... TRUE on success or FALSE on failure
History...... 26/03/07 - Initial version - LW
-------------------------------------------------------------------	*/
void SetSorterPaths(const DEVICE* dev, UBYTE coinposition, ULONG paths)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->set_sorter_paths(dev, coinposition, paths);
}

/*-------------------------------------------------------------------
Name......... EnableCoin()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Enable the inhibit for a specific coin
Parameters... 1) The device object passed in
............. 2) The coin denomination to enable
Return....... TRUE on success or FALSE on failure
History...... 26/03/07 - Initial version - LW
-------------------------------------------------------------------	*/
void EnableCoin(const DEVICE* dev, UBYTE coin)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->enable_coin(dev,coin);
}

/*-------------------------------------------------------------------
Name......... DisableCoin()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Disable a specific coin from being accepted
Parameters... 1) The device object passed in
............. 2) The specified coin denomination
Return....... TRUE on success or FALSE on failure
History...... 26/03/07 - Initial version - LW
-------------------------------------------------------------------	*/
void DisableCoin(const DEVICE* dev, UBYTE coin)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->disable_coin(dev,coin);
}

/*-------------------------------------------------------------------
Name......... EnableAllCoins()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Enable all the inhibits
Parameters... 1) The device object passed in
Return....... TRUE on success or FALSE on failure
History...... 26/03/07 - Initial version - LW
-------------------------------------------------------------------	*/
void EnableAllCoins(const DEVICE* dev)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->enable_all_coins(dev);
}

/*-------------------------------------------------------------------
Name......... DisableAllCoin()
Author....... Lee Weekes
Date......... 26/03/2007
Description.. Disable all inhibits for the coin acceptor
Parameters... 1) The device object passed in
Return....... TRUE on success or FALSE on failure
History...... 26/03/07 - Initial version - LW
-------------------------------------------------------------------	*/
void DisableAllCoins(const DEVICE* dev)
{
	const COINSERVICES* cser;

	cser=dev->services;

	cser->disable_all_coins(dev);
}
