/*-------------------------------------------------------------------
Name......... fat32.c
Author....... Mark Perkins
Date......... 20/10/02
Description.. Fat32 driver
-------------------------------------------------------------------*/

/*-------------------------------------------------------------------
DECLARATION FORMATS
Type defs........... UPPER CASE
#defines............ UPPER CASE
Global Functions.... GlobalFunction();
Static Functions.... staticfunction();
Global Variables.... GlobalVariable;
Static Variables.... staticvariable;
Local Variables..... localvariable;
Structure Elements.. Structure.element;
-------------------------------------------------------------------*/

/****	Include Files */

#include	"hardware.h"

#include	"fat32.h"
#include	"fat32.sp"

/****	Local Definitions */
int tolower(int c);
char *strcat(char *dst, const char *src);

/****	Global Variables */
static FILESYSTEM_DESCRIPTOR filesystem = {!MOUNTED,"\0",0,0,0,0,0,0,"\0"};
static DIR_DESCRIPTOR current_dir;
static SECTORDATA sectordata, clusterdata;
static ULONG current_sector=0;

/* error structure */
static	FAT32ERROR fat32error;

/* simple lockout flag

This flag is set during FAT32 API functions that call function ReadSectors().

The state of this flag is returned by API function FAT32LockoutStatus()

*/
static BYTE fat32_lockout;


/*
This variable was created in order to overcome the problem in the
FAT32 driver function load_file().
This function will load a whole number of sectors. Though imagine an image
800*600* (2bytes per pixel) + header = 960008 bytes on the compact flash.

If we load this from the CompactFlash, we need:
960008/512 (bytes per sector) = 1875'01 sectors.
As it's not an integer number, the number of sectors which will be read
from compact flash is 1875+1 => 1876, which is 512*1876 = 960512 bytes

Meaning that in video memory there are 960512-960008 bytes of non significant
data.

IN MOST OF THE CASES IT DOESN'T MATTER as customer will load images linearly
in video memory. Then they will update the offset according to the size of
their images. This way, they do not need to know how much data were loaded
and they automatically erase insignificant data from memory.

Though in some cases (see Q1156), customer allocate space in video memory
a section to store identical size images and they will use this space to
reload from CF some images if they need... This is when load_file will cause a
problem. Indeed this 504 extra bytes "loaded" will corrupt the next graphic..

This variable will be used to return the real_size loaded customers
identical circumstances will be told about this new function
 */

static ULONG real_size_loaded;


/****	Local Variables */

/****	Global Constants */

/****	Local Constants */

/************************************************************************
*****	Functions
*************************************************************************/

/*-------------------------------------------------------------------
Name......... GetDir()
Author....... Alan Austin
Date......... 1/11/02
Description.. Return current directory.
Parameters... 1) Address of device structure.
Return....... Return current directory.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE *GetDir(const DEVICE* dev)
{
	return(filesystem.current_path);
}

/*-------------------------------------------------------------------
Name......... FAT32GetStatus()
Author....... Alan Austin
Date......... 1/11/02
Description.. Return file system status.
Parameters... 1) Address of device structure.
Return....... Return status.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE FAT32GetStatus(const DEVICE* dev)
{
	return(filesystem.status);
}

/*-------------------------------------------------------------------
Name......... FAT32VolumeLabel()
Author....... Alan Austin
Date......... 1/11/02
Description.. Return volume label.
Parameters... 1) Address of device structure.
Return....... Return volume label.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE *FAT32VolumeLabel(const DEVICE* dev)
{
	return(filesystem.label);
}

/*-------------------------------------------------------------------
Name......... FAT32PartitionSize()
Author....... Alan Austin
Date......... 1/11/02
Description.. Return partition size.
Parameters... 1) Address of device structure.
Return....... Return partition size.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
ULONG FAT32PartitionSize(const DEVICE* dev)
{
	return(filesystem.partition_size);
}

/*-------------------------------------------------------------------
Name......... MountFAT32()
Author....... Alan Austin
Date......... 1/03/05
Description.. Read boot sector. Look for and mount FAT32 partition.
Parameters... 1) Address of device structure.
Return....... Return FAT32_OK on success.
............. Return fail code  on failure.
History...... 1/03/05 - Wrapper function for MountFAT32Filesystem()
............. for backwards compatibility - AA
-------------------------------------------------------------------*/
UBYTE MountFAT32(const DEVICE* dev)
{
	UBYTE error;

	error = MountFAT32Filesystem(dev, &fat32error);

	/* fatal errors only */
	return(error & FAT32_INVALID);
}


/*-------------------------------------------------------------------
Name......... MountFAT32Filesystem()
Author....... Alan Austin
Date......... 1 March 2005
Description.. Read boot sector. Look for and mount FAT32 partition.
Parameters... 1) Address of device structure.
............. 2) Pointer to error structure
Return....... FAT32_OK, FAT32_UNEXPECTED, FAT32_INVALID
History...... 1/03/05
............. Rework of old MountFAT32() function
.............
............. error handling/reporting improved
............. sector data buffer now local instead of global
............. AA
-------------------------------------------------------------------*/
UBYTE MountFAT32Filesystem(const DEVICE* dev, FAT32ERROR *error)
{
	SECTORDATA sector_data;
	ULONG partition_start_sector, partition_size;
	UBYTE i;
	UBYTE fail_code=0;

 	fat32_lockout = TRUE;	/* set lockout flag */

	/* clear error structure */
	error->result=FAT32_OK;
	for(i=0; i<FAT32_ERROR_END; i++)
		error->ident[i] = FAT32_OK;

	/* read boot sector */
	ReadSectors(&ATADevice, 0, sector_data.bw, 1);
	/* validate MBR */
	fail_code = check_valid_MBR(&sector_data.mbr, error);

	if(fail_code & FAT32_INVALID) {
		fat32_lockout = FALSE;	/* clear lockout flag */
		return(fail_code);
	}

	/*** workaround for mis-formatted CF cards without MBR ***/
	if(error->ident[MBR_ERROR_NO_PARTITION] == FAT32_UNEXPECTED)
		partition_start_sector = 0;
	else
		partition_start_sector = char2long(sector_data.mbr.partition1_sector_offset);

	partition_size = BYTES_PER_SECTOR*char2long(sector_data.mbr.partition1_sector_total);
	partition_size = partition_size>>20;	/* convert from bytes to megabytes */

	/* read BPM (BIOS Parameter Block) at start of FAT32 partition and initialise file system structure*/
	ReadSectors(&ATADevice, partition_start_sector, sector_data.bw, 1);
	fail_code = check_valid_BPB(&sector_data.bpb, error);

	fat32_lockout = FALSE;	/* clear lockout flag */

	/* validate BPM */
	if(fail_code & FAT32_INVALID) {
		return(fail_code);
	}

	/* initialise filesystem structure */
	strcpy(filesystem.current_path, "\\");
	filesystem.status = MOUNTED;
	for(i=0; i<11; i++)
	{
		filesystem.label[i] = sector_data.bpb.volume_label[i];
	}
	filesystem.label[11] = '\0';
	filesystem.partition_size			= partition_size;
	filesystem.sectors_per_cluster		= sector_data.bpb.sectors_per_cluster;
	filesystem.FAT_start_sector			= partition_start_sector + char2word(sector_data.bpb.reserved_sectors);
	filesystem.data_start_sector 		= filesystem.FAT_start_sector + sector_data.bpb.num_fats * char2long(sector_data.bpb.fat_size32);
	filesystem.root_dir_start_cluster 	= char2long(sector_data.bpb.root_start_cluster32);
	filesystem.current_dir_start_cluster= filesystem.root_dir_start_cluster;

	current_dir.next_cluster = filesystem.current_dir_start_cluster;
	current_dir.next_record = 0;

	return(fail_code);
}

/*-------------------------------------------------------------------
Name......... ChangeDir()
Author....... Alan Austin
Date......... 1/11/02
Description.. Change current directory (cf DOS 'cd' command).
Parameters... 1) Address of device structure.
............. 2) Pointer to path.
Return....... Return 0 on success.
............. Return 1 on failure.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE ChangeDir(const DEVICE* dev, BYTE *path)
{
	ULONG start_cluster;
	FILE_DESCRIPTOR file_desc;

	/* root directory */
	if(!strcmp(path, "\\"))
	{
		filesystem.current_dir_start_cluster = filesystem.root_dir_start_cluster;
		current_dir.next_cluster = filesystem.current_dir_start_cluster;
		current_dir.next_record = 0;
		strcpy(filesystem.current_path, "\\");
		return(0);
	}

	/* current directory */
	if(!strcmp(path, "."))
	{
		current_dir.next_cluster = filesystem.current_dir_start_cluster;
		current_dir.next_record = 0;
		return(0);
	}

	/* sub-directory */
	fat32_lockout = TRUE;	/* set lockout flag */
	if ((start_cluster = search_directory(dev, path, ATTR_DIRECTORY, &file_desc)))
	{
		filesystem.current_dir_start_cluster = start_cluster;
		current_dir.next_cluster = filesystem.current_dir_start_cluster;
		current_dir.next_record = 0;
		current_dir.next_cluster = filesystem.current_dir_start_cluster;
		current_dir.next_record = 0;

		/* parent directory */
		if(!strcmp(path, ".."))
		{
			strip_path(filesystem.current_path);
		}
		else
		{
			strcat(filesystem.current_path, path);
		}

		strcat(filesystem.current_path, "\\");
		fat32_lockout = FALSE;	/* clear lockout flag */
		return(0);
	}

	fat32_lockout = FALSE;	/* clear lockout flag */
	return(1);
}

/*-------------------------------------------------------------------
Name......... FAT32LoadFile()
Author....... Alan Austin
Date......... 1/11/02
Description.. Load file to memory.
Parameters... 1) Address of device structure.
............. 2) Pointer to path.
............. 3) Pointer to memory.
Return....... Return 0 on success.
............. Return 1 on failure.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE FAT32LoadFile(const DEVICE* dev, BYTE *path, UWORD *mem)
{
	ULONG start_cluster;
	FILE_DESCRIPTOR file_desc;

	fat32_lockout = TRUE;	/* set lockout flag */

	if ((start_cluster = search_directory(dev, path, 0, &file_desc)))
	{
		load_file(start_cluster, mem);
		fat32_lockout = FALSE;	/* clear lockout flag */
		return(0);
	}
	fat32_lockout = FALSE;	/* clear lockout flag */
	return(1);
}

/*-------------------------------------------------------------------
Name......... load_file()
Author....... Alan Austin
Date......... 1/11/02
Description.. Load file beginning at start cluster to memory.
Parameters... 1) Start cluster.
............. 2) Pointer to memory.
Return....... Return 0.
History...... 06/08/03 - First Pass  - AA
............. 01/03/05 - Unused size parameter removed - AA
-------------------------------------------------------------------*/
UBYTE load_file(ULONG start_cluster, UWORD *mem)
{
	RLE_CLUSTER_LIST *p, cluster_list;
	ULONG cluster;
	ULONG sector;
	ULONG n;

	UWORD *initial_mem;

	/* initialise the static variable related to real size loaded*/
	real_size_loaded=0;
	/* copy the content of the initial memory */
	initial_mem=mem;

	/* copy cluster list from drive */
	p=&cluster_list;
	copy_cluster_list(start_cluster, p);

	n=0;

	cluster = p->cluster_value[n];
	sector = filesystem.data_start_sector + (cluster-2)*filesystem.sectors_per_cluster;

	while (cluster < CLUSTER_EOC)
	{
		ReadSectors(&ATADevice, sector, mem, p->repeat_count[n]);

		if (p->repeat_count[n])
		{
			mem += ((BYTES_PER_SECTOR/2)*p->repeat_count[n]);
		}
		else
		{
			/* repeat_count=0 is 256 repeats */
			mem += ((BYTES_PER_SECTOR/2)*256);
		}

		n++;

		/* for each cluster in list... */
		cluster = p->cluster_value[n];
		sector = filesystem.data_start_sector + (cluster-2)*
				 filesystem.sectors_per_cluster;
	}

	/* update the number of bytes really loaded */
	real_size_loaded=(mem-initial_mem)*2;

	return(0);
}

/*-------------------------------------------------------------------
Name......... next_directory_record()
Author....... Alan Austin
Date......... 1/11/02
Description.. Get next directory record from directory.
Parameters... 1) Pointer to directory descriptor.
Return....... Return diretory enrty.
History...... 09/09/04 - Changed the CLUSTER_EOC check to >= to
..............avoid directory listing from hanging (port of fix from
..............Pluto 5 issue 8 library to Pluto 6) - EL
..............06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
static UBYTE *next_directory_record(DIR_DESCRIPTOR *dir)
{
	ULONG sector, record;

	record=dir->next_record;

	if (dir->next_cluster == CLUSTER_FREE)
	{
		return(NULL);
	}
	if (dir->next_cluster == CLUSTER_BAD)
	{
		return(NULL);
	}
	if (dir->next_cluster >= CLUSTER_EOC)
	{
		return(NULL);
	}

	/* get sector containing next record */
	sector = filesystem.data_start_sector + (dir->next_cluster - 2)*
			 filesystem.sectors_per_cluster	+ record/DIR_RECORDS_PER_SECTOR;

	ReadSectors(&ATADevice, sector, sectordata.bw, 1);

	/* increment to next record */
	dir->next_record++;

	/* start of next cluster? */
	if (dir->next_record==filesystem.sectors_per_cluster*DIR_RECORDS_PER_SECTOR)
	{
		dir->next_record = 0;
		dir->next_cluster = read_cluster_entry(dir->next_cluster);
	}
	return(&sectordata.bb[32*(record%DIR_RECORDS_PER_SECTOR)]);
}

/*-------------------------------------------------------------------
Name......... NextDirRecord()
Author....... Alan Austin
Date......... 1/11/02
Description.. Return next record from current directory
Parameters... 1) Address of device structure.
............. 2) Pointer to directory entry.
Return....... Return 0 when no entry.
............. Return 1 when entry found.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE NextDirRecord(const DEVICE* dev, DIR_ENTRY* dir_entry)
{
	UBYTE *directory_entry;
	UBYTE *long_filename, name_buffer[256];
	UBYTE alias_str[16];			/* 8.3 */
	UBYTE attrib_str[16];			/* <dir> rhs */

	fat32_lockout = TRUE;	/* set lockout flag */

	name_buffer[255] = '\0'; long_filename = &name_buffer[255];

	/* for each directory record... */
	while ( (directory_entry = next_directory_record(&current_dir)) != NULL )
	{
		switch (directory_entry[0])
		{
			case NULL_ENTRY:
				/* END OF RECORDS. Bail out... */
				fat32_lockout = FALSE;	/* clear lockout flag */
				return(0);
				break;

			case DELETED_ENTRY:
				/* don't print */
				break;

			default:
				if (directory_entry[ATTRIBUTE] == ATTR_LONG_NAME)
				{
					/* build up long file name... */
					long_filename=build_filename(directory_entry, long_filename);
				}
				else
				{
					/* fill in directory entry record */
					strcpy(dir_entry->alias,alias(directory_entry, alias_str));
					strcpy(dir_entry->long_filename,long_filename);
					strcpy(dir_entry->attrib_str,attrib(directory_entry, attrib_str));
					dir_entry->filesize = char2long(&directory_entry[FILESIZE]);
					fat32_lockout = FALSE;	/* clear lockout flag */
					return(1);
				}
				break;
		}
	}

	fat32_lockout = FALSE;	/* clear lockout flag */
	return(0);
}

/*-------------------------------------------------------------------
Name......... DirEntryAttrib()
Author....... Alan Austin
Date......... 1/11/02
Description.. Return directory entry attribute string: format  <dir> rhs.
Parameters... 1) Address of device structure.
............. 2) Pointer to directory entry.
Return....... Returns pointer to attrib.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE *DirEntryAttrib(const DEVICE* dev, DIR_ENTRY* dir_entry)
{
	return(dir_entry->attrib_str);
}

/*-------------------------------------------------------------------
Name......... DirEntryAlias()
Author....... Alan Austin
Date......... 1/11/02
Description.. Return directory entry alias (short filename)
............. string: 8.3 format.
Parameters... 1) Address of device structure.
............. 2) Pointer to directory entry.
Return....... Returns pointer to alias.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE *DirEntryAlias(const DEVICE* dev, DIR_ENTRY* dir_entry)
{
	return(dir_entry->alias);
}

/*-------------------------------------------------------------------
Name......... DirEntryFilesize()
Author....... Alan Austin
Date......... 1/11/02
Description.. Return directory entry file size.
Parameters... 1) Address of device structure.
............. 2) Pointer to directory entry.
Return....... Returns file size.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
ULONG DirEntryFilesize(const DEVICE* dev, DIR_ENTRY* dir_entry)
{
	return(dir_entry->filesize);
}

/*-------------------------------------------------------------------
Name......... DirEntryLongfilename()
Author....... Alan Austin
Date......... 1/11/02
Description.. Return directory entry long filename.
Parameters... 1) Address of device structure.
............. 2) Pointer to directory entry.
Return....... Returns pointer to long filename
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE *DirEntryLongfilename(const DEVICE* dev, DIR_ENTRY* dir_entry)
{
	return(dir_entry->long_filename);

}

/*-------------------------------------------------------------------
Name......... search_directory()
Author....... Alan Austin
Date......... 1/11/02
Description.. search for file matching path and attributes
............. in current directory
............. Sequential search through directory records.
............. There is no alternative!
Parameters... 1) Address of device structure.
............. 2) Pointer to path.
............. 3) Attributes.
............. 4) Pointer to file descriptor.
Return....... Returns 0 on failure
............. Returns start cluster on success
History...... 22/10/04 - remove call to ChangeDir to save stack space
............. 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
static ULONG search_directory(const DEVICE* dev, UBYTE *path,
							  UBYTE attributes,
							  FILE_DESCRIPTOR *file_desc)
{
	UBYTE *directory_entry;
	UBYTE *long_filename, name_buffer[256];
	UBYTE alias_str[16];			/* 8.3 */

	file_desc->start_cluster = 0;
	file_desc->size = 0;

	/* bail out for empty strings */
	if(path[0]=='\0')
	{
		return(0);
	}

	name_buffer[255] = '\0'; long_filename = &name_buffer[255];

	/* initialise current directory */
	current_dir.next_cluster = filesystem.current_dir_start_cluster;
	current_dir.next_record = 0;

	/* for each directory record...*/
	while ( (directory_entry = next_directory_record(&current_dir)) != NULL )
	{
		switch (directory_entry[0])
		{
			case NULL_ENTRY:
				/* END OF RECORDS. Bail out... */
				return(0);
				break;

			case DELETED_ENTRY:
				/* don't print */
				break;

			default:
				if (directory_entry[ATTRIBUTE] == ATTR_LONG_NAME)
				{
					/* build up long file name... */
					long_filename=build_filename(directory_entry, long_filename);
				}
				else
				{
					/* COMPARE... */
					if ( (!_strcmp(path, alias(directory_entry, alias_str), CASE_SENSITIVITY)) ||
						 (!_strcmp(path, long_filename, CASE_SENSITIVITY))  )
					{
						 /* name match! */
						 /* now check attributes... */
						 /* (INITIAL VERSION check only directory attribute */

						 if ((ATTR_DIRECTORY & attributes  ) ==
						 	 (ATTR_DIRECTORY & directory_entry[ATTRIBUTE]))
						 {

						 	/* success! */
							file_desc->start_cluster =
								 (ULONG) ((char2word(&directory_entry[START_CLUSTER_HI])) << 16)
								+(ULONG) char2word(&directory_entry[START_CLUSTER_LO]);

							file_desc->size = char2long(&directory_entry[FILESIZE]);

							/* special case: dot-dot entry in sub-directory of root directory: */
							if ( (file_desc->start_cluster == 0) && (!_strcmp(path, "..", CASE_SENSITIVITY)))
							{
								file_desc->start_cluster = filesystem.root_dir_start_cluster;
							}

						 	return(file_desc->start_cluster);
						 }
					}

					name_buffer[255] = '\0'; long_filename = &name_buffer[255];
				}
				break;
		}
	}

	/* FAILED MATCH */
	return(0);
}

/*-------------------------------------------------------------------
Name......... copy_cluster_list()
Author....... Alan Austin
Date......... 1/11/02
Description.. Read cluster linked list headed by start_cluster from drive FAT
............. create run length encoded local copy of list in RAM.
Parameters... 1) Start cluster.
............. 2) Pointer to cluster list.
Return....... Returns 0
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
static UBYTE copy_cluster_list(ULONG start_cluster, RLE_CLUSTER_LIST *p)
{
	ULONG n;
	ULONG cluster_value;
	ULONG next_cluster_value;

	n = 0;
	p->repeat_count[n] = filesystem.sectors_per_cluster;
	p->cluster_value[n] = start_cluster;
	next_cluster_value = start_cluster;

	do {
		cluster_value = read_cluster_entry(next_cluster_value);

		if (cluster_value == CLUSTER_FREE)
		{
			p->cluster_value[0] = CLUSTER_EOC;
			return(1);
		}
		if (cluster_value == CLUSTER_BAD)
		{
			p->cluster_value[0] = CLUSTER_EOC;
			return(2);
		}

		/* contiguous clusters: increment repeat count
		repeat count is unsigned byte so wrap round to 0 at max count of 256
		*/
		if ((cluster_value==next_cluster_value+1) && (p->repeat_count[n] != 0))
		{
			p->repeat_count[n] += filesystem.sectors_per_cluster;
		}
		else
		{
			n++;
			p->cluster_value[n] = cluster_value;
			p->repeat_count[n] = filesystem.sectors_per_cluster;
		}

		next_cluster_value = cluster_value;

	} while ((cluster_value < CLUSTER_EOC) && (n < MAX_CLUSTERS));

	p->cluster_value[MAX_CLUSTERS-1] = CLUSTER_EOC;	/* precaution */

	return(0);
}

/*-------------------------------------------------------------------
Name......... read_cluster_entry()
Author....... Alan Austin
Date......... 1/11/02
Description.. read FAT and return cluster entry for given cluster number.
Parameters... 1) Cluster Number.
Return....... Cluster Entry.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
static ULONG read_cluster_entry(ULONG cluster_number)
{
	ULONG sector;
	ULONG offset;

	/*
	NOTE: hard wired value for cluster->sector conversion
	1 cluster entry = 4 bytes, 1 sector = 512bytes
	*/

	sector = filesystem.FAT_start_sector + cluster_number*4/BYTES_PER_SECTOR;
	offset = (cluster_number*4)% BYTES_PER_SECTOR;

	if (sector != current_sector)
	{
		ReadSectors(&ATADevice, sector, clusterdata.bw, 1);
	}
	current_sector = sector;
	return(char2long(&clusterdata.bb[offset]));
}

/*-------------------------------------------------------------------
Name......... alias()
Author....... Alan Austin
Date......... 1/11/02
Description.. Extract alias filename from directory entry
Parameters... 1) Pointer to entry.
............. 2) Pointer to alias string.
Return....... Pointer to alias.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
static UBYTE *alias(UBYTE *entry, UBYTE *alias_str)
{
	LONG i;
	UBYTE *p;

	p=alias_str;
	/* filename */
	for(i=0; i<8; i++, entry++)
	{
		if (*entry != ' ')
		{
			*p++ = *entry;
		}
	}
	/* extension */
	if ((*entry != ' ') && (*entry != '-'))
	{
		*p++ = '.';
		for(i=0; i<3; i++, entry++)
		{
			if (*entry != ' ')
			{
				*p++ = *entry;
			}
		}
	}

	/* null termination */
	*p = '\0';

	return(alias_str);

}

/*-------------------------------------------------------------------
Name......... attrib()
Author....... Alan Austin
Date......... 1/11/02
Description.. Extract attribute string from directory entry
Parameters... 1) Pointer to entry.
............. 2) Pointer to attribute string.
Return....... Pointer to attrib.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
static UBYTE *attrib(UBYTE *entry, UBYTE *attrib_str)
{
	UBYTE *p;

	p=attrib_str;

	if (entry[ATTRIBUTE] & ATTR_DIRECTORY)
	{
		strcpy(p, "<dir>    ");
	}
	else
	{
		strcpy(p, "         ");
	}

	if (entry[ATTRIBUTE] & ATTR_READ_ONLY)
	{
		*(p+6) = 'r';
	}
	if (entry[ATTRIBUTE] & ATTR_HIDDEN)
	{
		*(p+7) = 's';
	}
	if (entry[ATTRIBUTE] & ATTR_SYSTEM)
	{
		*(p+8) = 'h';
	}

	/* null termination */
	*(p+9)= '\0';

	return(attrib_str);
}

/*-------------------------------------------------------------------
Name......... buildfilename()
Author....... Alan Austin
Date......... 1/11/02
Description.. Build up long filename over multiple function calls
............. from sequence of slot type  directory entries
Parameters... 1) Pointer to string.
............. 2) Pointer to string.
Return....... Pointer to name.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
static UBYTE *build_filename(UBYTE *entry, UBYTE *name)
{

	*(--name) = entry[NAME13];
	*(--name) = entry[NAME12];
	*(--name) = entry[NAME11];
	*(--name) = entry[NAME10];
	*(--name) = entry[NAME9];
	*(--name) = entry[NAME8];
	*(--name) = entry[NAME7];
	*(--name) = entry[NAME6];
	*(--name) = entry[NAME5];
	*(--name) = entry[NAME4];
	*(--name) = entry[NAME3];
	*(--name) = entry[NAME2];
	*(--name) = entry[NAME1];

	return(name);
}

/*-------------------------------------------------------------------
Name......... _strcmp()
Author....... Alan Austin
Date......... 1/11/02
Description.. String compare function with case sensitivity.
Parameters... 1) Pointer to string.
............. 2) Pointer to string.
............. 3) Case sensitive option.
Return....... Returns 0 if successfull.
............. Returns positive value if failed.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
static BYTE _strcmp(BYTE *s, BYTE *t, BYTE case_sensitive)
{

	if (case_sensitive)
	{
		for(; *s == *t; s++, t++)
		{
			if (*s == '\0')
			{
				return(0);
			}
		}
	}
	else
	{
		for(; tolower(*s) == tolower(*t); s++, t++)
		{
			if (*s == '\0')
			{
				return(0);
			}
		}
	}
	return(*s-*t);
}

/*-------------------------------------------------------------------
Name......... _strncmp()
Author....... Alan Austin
Date......... 21 April 2005
Description.. String compare function with max length.
Parameters... 1) Pointer to string.
............. 2) Pointer to string.
............. 3) Length.
Return....... Returns 0 if successfull.
............. Returns positive value if failed.
History...... 21 April 2005 - First Pass  - AA
-------------------------------------------------------------------*/
static BYTE _strncmp(BYTE *s, BYTE *t, UWORD maxchars)
{

	UWORD n;

	for(n=0; *s == *t; s++, t++)
	{
		if (++n >= maxchars)
		{
			return(0);
		}
	}

	return(*s-*t);
}



/*-------------------------------------------------------------------
Name......... char2word()
Author....... Alan Austin
Date......... 1/11/02
Description.. Char to word conversion.
Parameters... 1) pointer to char.
Return....... None.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
static UWORD char2word(UBYTE *p)
{
	return((UWORD)(*p + ((*(p+1))<<8) ));
}

/*-------------------------------------------------------------------
Name......... char2long()
Author....... Alan Austin
Date......... 1/11/02
Description.. Char to long conversion.
Parameters... 1) pointer to char.
Return....... None.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
static ULONG char2long(UBYTE *p)
{
	return((ULONG)(*p + ((*(p+1))<<8) + ((*(p+2))<<16) + ((*(p+3))<<24) ));
}

/*-------------------------------------------------------------------
Name......... strip_path()
Author....... Alan Austin
Date......... 1/11/02
Description.. Strip current directory from path string.
Parameters... 1) pointer to string.
Return....... Returns 0 if successfull.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
static BYTE strip_path(UBYTE *p)
{
	while (*p != '\0')
	{
		p++;
	}
	while (*p != '\\')
	{
		p--;
	}
	p--;
	while (*p != '\\')
	{
		p--;
	}
	*p = '\0';

	return(0);
}

/*-------------------------------------------------------------------
Name......... FileOpen()
Author....... Alan Austin
Date......... 06/08/03
Description.. Open file from ATA device.
............. Initialise file pointer structure FILE_DESCRIPTOR.
Parameters... 1) Address of device structure.
............. 2) Pointer to filename.
............. 3) Pointer to file descriptor.
............. 4) Mode.
Return....... Returns poitive value on success.
............. Returns NULL pointer on failure.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
FILE_DESCRIPTOR 	*FileOpen(const DEVICE* dev, BYTE *filename,
							  FILE_DESCRIPTOR *file, UBYTE mode)
{
	fat32_lockout = TRUE;	/* set lockout flag */

	/* set file descriptor fields: start_cluster, filesize */
	if (!search_directory(dev, filename, 0, file))
	{
		file->status = FILE_CLOSED;
		fat32_lockout = FALSE;	/* clear lockout flag */
		return(NULL);
	}

	/* copy cluster list from drive to file descriptor */
	copy_cluster_list(file->start_cluster, &file->cluster_list);

	file->cluster_list.index = 0;
	file->cluster_list.offset = 0;
	file->current_byte=0;
	file->current_sector=0;
	file->status = FILE_OPEN;
	file->flags = mode;

	fat32_lockout = FALSE;	/* clear lockout flag */
	return(file);
}

/*-------------------------------------------------------------------
Name......... FileGetChar()
Author....... Alan Austin
Date......... 06/08/03
Description.. File Character Read.
Parameters... 1) Address of device structure.
............. 2) Pointer to file descriptor.
Return....... Returns next char from file on success.
............. Returns EOF on fail.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UWORD FileGetChar(const DEVICE* dev, FILE_DESCRIPTOR *file)
{
	ULONG cluster;
	ULONG sector;
	UWORD ch;

	/* file not open */
	if (file->status == FILE_CLOSED) {
		return(EOF);
	}

	fat32_lockout = TRUE;	/* set lockout flag */

	/* has the the sector containing the next byte been loaded? */

	if ((file->current_byte % BYTES_PER_SECTOR) == 0)
	{
		/* reached end of sector: load next sector */
		if ((cluster = file->cluster_list.cluster_value[file->cluster_list.index]) < CLUSTER_EOC)
		{
			sector = filesystem.data_start_sector + (cluster-2)*filesystem.sectors_per_cluster;
			sector += (ULONG) file->cluster_list.offset;

			ReadSectors(&ATADevice, sector, file->buffer.bw, 1);

			file->cluster_list.offset++;
			if (file->cluster_list.offset == file->cluster_list.repeat_count[file->cluster_list.index])
			{
				file->cluster_list.offset=0;
				file->cluster_list.index++;
			}


			file->current_sector++;
		}
	}

	fat32_lockout = FALSE;	/* clear lockout flag */

	ch = file->buffer.bb[file->current_byte % BYTES_PER_SECTOR];

	file->current_byte++;
	if (file->current_byte <= file->size)
	{
		return(ch);
	}
	else
	{
		return(EOF);
	}
}

/*-------------------------------------------------------------------
Name......... FileGetSectors()
Author....... Alan Austin
Date......... 06/08/03
Description.. File Block Read.
............. Read n sectors of data from file to buffer at current
............. reading position.
............. Returns actual number of sectors read.
Parameters... 1) Address of device structure.
............. 2) Pointer to file descriptor.
............. 3) Pointer to memory to output data.
............. 4) Number of sectors to read.
Return....... Returns sectors loaded.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
ULONG FileGetSectors(const DEVICE* dev, FILE_DESCRIPTOR *file,
					 UWORD *mem, ULONG nsectors)
{
	ULONG cluster;
	ULONG sector;
	UWORD repeat_count;
	ULONG sectors_loaded;

	sectors_loaded=0;

	/* file not open */
	if (file->status == FILE_CLOSED)
	{
		return(sectors_loaded);
	}

	fat32_lockout = TRUE;	/* set lockout flag */
	while (	nsectors &&
			((cluster = file->cluster_list.cluster_value[file->cluster_list.index]) < CLUSTER_EOC))
	{
		sector = filesystem.data_start_sector + (cluster-2)*filesystem.sectors_per_cluster;

		sector += (ULONG) file->cluster_list.offset;

		repeat_count = (UWORD) (file->cluster_list.repeat_count[file->cluster_list.index]);
		if(repeat_count == 0)
		{
			repeat_count = 256;	/* special case */
		}
		repeat_count -= (UWORD) file->cluster_list.offset;

		if(nsectors < repeat_count)
		{
			repeat_count = nsectors;
			file->cluster_list.offset += repeat_count;
		}
		else
		{
			file->cluster_list.offset = 0;
			file->cluster_list.index++;
		}

		ReadSectors(&ATADevice, sector, mem, (UBYTE) repeat_count);

		mem += ((BYTES_PER_SECTOR/2)*repeat_count);

		nsectors -= repeat_count;
		sectors_loaded += repeat_count;

		file->current_sector += repeat_count;
		file->current_byte   =  BYTES_PER_SECTOR*file->current_sector;
	}

	fat32_lockout = FALSE;	/* clear lockout flag */
	return(sectors_loaded);
}

/*-------------------------------------------------------------------
Name......... FilePutSector()
Author....... Alan Austin
Date......... 12 May 2005
Description.. File Block Write.
............. Write sector of data from buffer to file at current
............. file position. Increments current file position
.............
Parameters... 1) Address of device structure.
............. 2) Pointer to file descriptor.
............. 3) Pointer to memory sector.
Return....... Returns non-zero on failure.
History...... 12 May 2005 - First Pass  - AA
-------------------------------------------------------------------*/
ULONG FilePutSector(const DEVICE* dev, FILE_DESCRIPTOR *file,
					 UWORD *mem)
{
	ULONG cluster;
	ULONG sector;
	UWORD repeat_count;
	ULONG file_last_sector;


	file_last_sector = (file->size - 1)/BYTES_PER_SECTOR;
	/* sector out of range error */
	if(file->current_sector > file_last_sector)
		return(FILE_SECTOR_OUT_OF_RANGE);

	/* file not open error */
	if(file->status == FILE_CLOSED)
		return(FILE_NOT_OPEN);

	/* file read only error */
	if(file->flags == READ_ONLY)
		return(FILE_READ_ONLY);


	fat32_lockout = TRUE;	/* set lockout flag */
	if ((cluster = file->cluster_list.cluster_value[file->cluster_list.index]) < CLUSTER_EOC) {

		sector = filesystem.data_start_sector + (cluster-2)*filesystem.sectors_per_cluster;

		sector += (ULONG) file->cluster_list.offset;

		repeat_count = (UWORD) (file->cluster_list.repeat_count[file->cluster_list.index]);
		if(repeat_count == 0)
		{
			repeat_count = 256;	/* special case */
		}
		repeat_count -= (UWORD) file->cluster_list.offset;

		if(repeat_count>1)
		{
			file->cluster_list.offset++;
		}
		else
		{
			file->cluster_list.offset = 0;
			file->cluster_list.index++;
		}

		WriteSector(&ATADevice, sector, mem);

		file->current_sector++;
		file->current_byte = BYTES_PER_SECTOR*file->current_sector;
	}

	fat32_lockout = FALSE;	/* clear lockout flag */
	return(0);
}

/*-------------------------------------------------------------------
Name......... FileLoad()
Author....... Alan Austin
Date......... 06/08/03
Description.. Load entire file to buffer.
............. Rewinds to start of file on completion.
Parameters... 1) Address of device structure.
............. 2) Pointer to file descriptor.
............. 2) Pointer to memory to output data.
Return....... Returns 0 if successfull.
............. Returns positive value if failed.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE FileLoad(const DEVICE* dev, FILE_DESCRIPTOR *file, UWORD *mem)
{
	ULONG cluster;
	ULONG sector;
	UWORD repeat_count;

	/* file not open */
	if (file->status == FILE_CLOSED) {
		return(1);
	}

	fat32_lockout = TRUE;	/* set lockout flag */
	file->cluster_list.index=0;

	while ((cluster = file->cluster_list.cluster_value[file->cluster_list.index]) < CLUSTER_EOC)
	{
		sector = filesystem.data_start_sector + (cluster-2)*filesystem.sectors_per_cluster;
		repeat_count = (UWORD) (file->cluster_list.repeat_count[file->cluster_list.index]);

		if(repeat_count == 0)
		{
			repeat_count = 256;	/* special case */
		}

		ReadSectors(&ATADevice, sector, mem, (UBYTE) repeat_count);

		mem += ((BYTES_PER_SECTOR/2)*repeat_count);

		file->cluster_list.index++;

	}

	FileRewind(file);
	fat32_lockout = FALSE;	/* clear lockout flag */
	return(0);
}

/*-------------------------------------------------------------------
Name......... FileRewind()
Author....... Alan Austin
Date......... 06/08/03
Description.. Rewind current reading position to start of file.
Parameters... 1) Pointer to file descriptor.
Return....... Returns 0 if successfull.
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE FileRewind(FILE_DESCRIPTOR *file)
{
	file->cluster_list.index = 0;
	file->cluster_list.offset = 0;

	file->current_byte=0;
	file->current_sector=0;

	return(0);
}

/*-------------------------------------------------------------------
Name......... FileSeekSector()
Author....... Alan Austin
Date......... 19 May 2005
Description.. File Sector Read.
............. Moves file current reading position
............. to sector offset
............. First sector is sector 0
.............
Parameters... 1) Address of device structure.
............. 2) Pointer to file descriptor.
............. 3) Seek sector number (offset from start of file)
Return....... Returns zero if successful
History...... 19 May 2005 - First Pass  - AA
-------------------------------------------------------------------*/
ULONG FileSeekSector(const DEVICE* dev, FILE_DESCRIPTOR *file, ULONG sector_offset)
{
	ULONG cluster;
	ULONG sector;
	UWORD repeat_count;
	ULONG file_last_sector;

	/* file not open error */
	if (file->status == FILE_CLOSED)
	{
		return(FILE_NOT_OPEN);
	}

	file_last_sector = (file->size - 1)/BYTES_PER_SECTOR;
	/* sector out of range error */
	if(sector_offset > file_last_sector)
		return(FILE_SECTOR_OUT_OF_RANGE);

	fat32_lockout = TRUE;	/* set lockout flag */

	/* Rewind current reading position to start of file. */
	file->cluster_list.index = 0;
	file->cluster_list.offset = 0;

	file->current_byte=0;
	file->current_sector=0;

	while (	sector_offset &&
			((cluster = file->cluster_list.cluster_value[file->cluster_list.index]) < CLUSTER_EOC))
	{
		sector = filesystem.data_start_sector + (cluster-2)*filesystem.sectors_per_cluster;

		sector += (ULONG) file->cluster_list.offset;

		repeat_count = (UWORD) (file->cluster_list.repeat_count[file->cluster_list.index]);
		if(repeat_count == 0)
		{
			repeat_count = 256;	/* special case */
		}
		repeat_count -= (UWORD) file->cluster_list.offset;

		if(sector_offset < repeat_count)
		{
			repeat_count = sector_offset;
			file->cluster_list.offset += repeat_count;
		}
		else
		{
			file->cluster_list.offset = 0;
			file->cluster_list.index++;
		}

		sector_offset -= repeat_count;

		file->current_sector += repeat_count;
		file->current_byte   =  BYTES_PER_SECTOR*file->current_sector;
	}

	fat32_lockout = FALSE;	/* clear lockout flag */
	return(0);
}

/*-------------------------------------------------------------------
Name......... FileCurrentSector()
Author....... Alan Austin
Date......... 19 May 2005
Description.. Returns file current r/w position
............. (the offset in sectors from start of file)
............. First sector is sector 0
.............
Parameters... 1) Address of device structure.
............. 2) Pointer to file descriptor.
Return....... Returns current position
History...... 19 May 2005 - First Pass  - AA
-------------------------------------------------------------------*/
ULONG FileCurrentSector(const DEVICE* dev, FILE_DESCRIPTOR *file)
{
	return(file->current_sector);
}


/*-------------------------------------------------------------------
Name......... FileClose()
Author....... Alan Austin
Date......... 06/08/03
Description.. Close the file pointed to by file descriptor.
Parameters... 1) Pointer to file descriptor.
Return....... Returns zero if successful
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
UBYTE FileClose (FILE_DESCRIPTOR *file)
{
	file->cluster_list.index = 0;
	file->cluster_list.offset = 0;

	file->current_byte=0;
	file->current_sector=0;

	file->status = FILE_CLOSED;

	return(0);
}

/*-------------------------------------------------------------------
Name......... GetFileSize()
Author....... Alan Austin
Date......... 06/08/03
Description.. Returns file size in bytes.
Parameters... 1) Pointer to file descriptor.
Return....... None
History...... 06/08/03 - First Pass  - AA
-------------------------------------------------------------------*/
ULONG GetFileSize(FILE_DESCRIPTOR *file)
{
	return(file->size);
}

/*-----------------------------------------------------------------------
Name......... check_valid_MBR
Author....... Alan Austin
Date......... 09 March 2005
Description.. Validate Master Boot Record
Parameters... 1) Pointer to MBR structure
............. 2) Pointer to error structure
Return....... FAT32_OK, FAT32_UNEXPECTED, FAT32_INVALID
Notes........
............. Sequence of tests on MBR values
.............
............. Errors classed either as:
.............   UNEXPECTED (non-fatal) or
.............   INVALID (fatal)
.............
History...... 09 March 2005 - First pass - AA
------------------------------------------------------------------------*/

static FAT32STATUS check_valid_MBR(MBR *mbr, FAT32ERROR *error)
{
	/* check partition active flag */
	if (mbr->partition1_state != PARTITION_ACTIVE)
	{
		error->ident[MBR_ERROR_NO_PARTITION] = FAT32_UNEXPECTED;
		error->result |= FAT32_UNEXPECTED;

		/**** "missing MBR" work-around: exit at this point ****/
		return(error->result);
	}

	/* check signature */
	if ((mbr->signature[1] != MBR_SIGNATURE_1) || (mbr->signature[0] != MBR_SIGNATURE_0))
	{
		error->ident[MBR_ERROR_SIGNATURE] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	/* check primary partition file system type */
	if ((mbr->partition1_type != FILESYSTEM_FAT32) && (mbr->partition1_type != FILESYSTEM_FAT32_NOTLBA))
	{
		error->ident[MBR_ERROR_FS_TYPE] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	return(error->result);
}

/*-----------------------------------------------------------------------
Name......... check_valid_BPB
Author....... Alan Austin
Date......... 09 March 2005
Description.. Validate Partition Bios Parameter Block
Parameters... 1) Pointer to BPB structure
............. 2) Pointer to error structure
Return....... FAT32_OK, FAT32_UNEXPECTED, FAT32_INVALID
Notes........
............. Sequence of tests on fields within BPB
............. Results placed in error structure members
.............
............. Errors are defined as:
.............   UNEXPECTED (non-fatal) or
.............   INVALID (fatal)
.............
............. This function is full of "magic" numbers defining allowed
............. values, ranges, strings
.............
............. The BPB is a jumble of fields of different fields,
............. with different data types: bytes, strings, longs
............. with a variety of different criterion for validity:
............. discrete value match, range match, match to multiple
............. discrete values etc
.............
.............
History...... 09 March 2005 - First pass - AA
------------------------------- -----------------------------------------*/

static FAT32STATUS check_valid_BPB(BPB *bpb, FAT32ERROR *error)
{
	/* jmp_boot[3]	Allowed: {0xeb,0x??,0x90} or {0xe9,0x??,0x??}*/
	if((bpb->jmp_boot[0]!=0xeb) && (bpb->jmp_boot[0]!=0xe9)){
		error->ident[BPB_ERROR_JMP_BOOT] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	/* oem_name[8]	Allowed: arbritrary name string. Expected: "MSDOS*.*" */
	if(_strncmp(bpb->oem_name, "MSDOS", 5)) {
		error->ident[BPB_ERROR_OEM_NAME] = FAT32_UNEXPECTED;
		error->result |= FAT32_UNEXPECTED;
	}

	/* bytes_per_sector[2]	Allowed: 512,1024,2048,4096. Expected: 512 (Compact Flash)*/
	switch(bpb->bytes_per_sector[1]) {
		case 512>>8:
			break;

		case 1024>>8:
		case 2048>>8:
		case 4096>>8:
			error->ident[BPB_ERROR_BYTES_PER_SECTOR] = FAT32_UNEXPECTED;
			error->result |= FAT32_UNEXPECTED;
			break;

		default:
			error->ident[BPB_ERROR_BYTES_PER_SECTOR] = FAT32_INVALID;
			error->result |= FAT32_INVALID;
			break;
	}

	/* sectors_per_cluster	Allowed: 1,2,4,8,16,32,64,128*/
	switch(bpb->sectors_per_cluster) {
		case 1:
		case 2:
		case 4:
		case 8:
		case 16:
		case 32:
		case 64:
		case 128:
			break;

		default:
			error->ident[BPB_ERROR_SECTORS_PER_CLUSTER] = FAT32_INVALID;
			error->result |= FAT32_INVALID;
			break;
	}

	/* reserved_sectors[2]	Allowed: non-zero. Expected: 32*/
	if(bpb->reserved_sectors[1] == 0)
	{
		if(bpb->reserved_sectors[0] == 0)
		{
			error->ident[BPB_ERROR_RESERVED_SECTORS] = FAT32_INVALID;
			error->result |= FAT32_INVALID;
		}
		else if (bpb->reserved_sectors[0] != 32)
		{
			error->ident[BPB_ERROR_RESERVED_SECTORS] = FAT32_UNEXPECTED;
			error->result |= FAT32_UNEXPECTED;
		}
	}


	 /* num_fats	Allowed: non-zero. Expected: 2*/
	if(bpb->num_fats == 0)
	{
		error->ident[BPB_ERROR_NUM_FATS] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	} else if(bpb->num_fats != 2)
	{
		error->ident[BPB_ERROR_NUM_FATS] = FAT32_UNEXPECTED;
		error->result |= FAT32_UNEXPECTED;
	}

	/* root_entries16[2]	Allowed: zero (FAT32)*/
	if(bpb->root_entries16[0] || bpb->root_entries16[1])
	{
		error->ident[BPB_ERROR_ROOT_ENTRIES16] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	/* total_sectors16[2]	Allowed: zero (FAT32)*/
	if(bpb->total_sectors16[0] || bpb->total_sectors16[1])
	{
		error->ident[BPB_ERROR_TOTAL_SECTORS16] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	/* media_descriptor	Allowed: 0xf0,0xf8-0xff. Expected: 0xf0 (removable) or 0xf8 (fixed)*/
	if((bpb->media_descriptor != 0xf0) && (bpb->media_descriptor != 0xf8))
	{
		if(bpb->media_descriptor > 0xf8)
		{
			error->ident[BPB_ERROR_MEDIA_DESCRIPTOR] = FAT32_UNEXPECTED;
			error->result |= FAT32_UNEXPECTED;
		}
		else
		{
			error->ident[BPB_ERROR_MEDIA_DESCRIPTOR] = FAT32_INVALID;
			error->result |= FAT32_INVALID;
		}
	}

	/* fat_size16[2]	Allowed: zero (FAT32)*/
	if(bpb->fat_size16[0] || bpb->fat_size16[1])
	{
		error->ident[BPB_ERROR_FAT_SIZE16] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	/* hidden_sectors[4]	Allowed: OS specific. Expected: zero (for non-partitioned)*/
	if(bpb->hidden_sectors[0] || bpb->hidden_sectors[1] || bpb->hidden_sectors[2] || bpb->hidden_sectors[3])
	{
		error->ident[BPB_ERROR_HIDDEN_SECTORS] = FAT32_UNEXPECTED;
		error->result |= FAT32_UNEXPECTED;
	}

	/* total_sectors32[4]	Allowed: non-zero*/
	if(!bpb->total_sectors32[0] && !bpb->total_sectors32[1] && !bpb->total_sectors32[2] && !bpb->total_sectors32[3])
	{
		error->ident[BPB_ERROR_TOTAL_SECTORS32] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	/* fat_size32[4]	Allowed: non-zero*/
	if(!bpb->fat_size32[0] && !bpb->fat_size32[1] && !bpb->fat_size32[2] && !bpb->fat_size32[3])
	{
		error->ident[BPB_ERROR_FAT_SIZE32] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	/* root_start_cluster32[4]	Allowed: non-zero. Expected: 2*/
	if(!bpb->root_start_cluster32[0] && !bpb->root_start_cluster32[1] && !bpb->root_start_cluster32[2] && !bpb->root_start_cluster32[3])
	{
		error->ident[BPB_ERROR_ROOT_START_CLUSTER32] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	/* fsinfo_start_sector32[2]	Allowed: non-zero. Expected: 1*/
	if(!bpb->fsinfo_start_sector32[0] && !bpb->fsinfo_start_sector32[1])
	{
		error->ident[BPB_ERROR_FSINFO_START_SECTOR32] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	/* backup_boot_sector[2]	Allowed: non-zero. Expected: 6*/
	if(!bpb->backup_boot_sector[0] && !bpb->backup_boot_sector[1])
	{
		error->ident[BPB_ERROR_BACKUP_BOOT_SECTOR] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	/* boot_signature	Allowed: 0x29*/
	if(bpb->boot_signature != 0x29)
	{
		error->ident[BPB_ERROR_BOOT_SIGNATURE] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	/* file_system_type[8]	Allowed: "FAT32  "*/
	if(_strncmp(bpb->file_system_type, "FAT32", 5))
	{
		error->ident[BPB_ERROR_FILE_SYSTEM_TYPE] = FAT32_INVALID;
		error->result |= FAT32_INVALID;
	}

	return(error->result);
}

/*-----------------------------------------------------------------------
Name......... FAT32LockoutStatus
Autho r....... Alan Austin
Date......... 8 April 2005
Description.. Returns current value of fat32_lockout flag
Parameters...
Return.......
History...... Introduced 8/04/05 AA
------------------------------------------------------------------------*/

BYTE FAT32LockoutStatus()
{
	return(fat32_lockout);
}

/*-----------------------------------------------------------------------
Name......... GetFileSizeLoaded()
Author....... Aymeric Dupont
Date......... 31/01/06
Description.. Returns number of bytes that have just been loaded by last
............. load_file call.
Parameters... None
Return....... Number of bytes that have just been loaded by last
............. load_file call
History...... First release 31/01/06 - AD
-------------------------------------------------------------------------*/
ULONG GetRealFileSizeLoaded()
{
	return(real_size_loaded);
}
