/*-------------------------------------------------------------------
Name......... mswinbmp.c
Author....... Mark Perkis
Date......... 11/09/03
Description.. Microsoft Windows BMP file loader.

Modified 4 January 2005 Alan Austin
readbmpdata24() modified to fix colour 0 transparency problem

-------------------------------------------------------------------*/

/*-------------------------------------------------------------------
DECLARATION FORMATS
Type defs........... UPPER CASE
#defines............ UPPER CASE
Global Functions.... GlobalFunction();
Static Functions.... staticfunction();
Global Variables.... GlobalVariable;
Static Variables.... staticvariable;
Local Variables..... localvariable;
Structure Elements.. Structure.element;
-------------------------------------------------------------------*/

/****	Include Files */
#include	"hardware.h"
#include	"interface.h"
#include	"mswinbmp.sp"

/****	Externals */

/****	Local Definitions */
#define ALPHABLENDING_MASK 		0x8000	/* set MSB of any pixel to 1 */
#define NONE_ALPHABLENDING_MASK 0x0000	/* set MSB of any pixel to 0 */
/****	Global Variables */

/****	Local Variables */
static UBYTE load_alphablended = FALSE;
/****	Global Constants */

/****	Local Constants */

/********************************************************************
*****	Functions
********************************************************************/
/*-------------------------------------------------------------------
Name......... LoadMsWinBmp24()
Author....... Mark Perkins
Date......... 17/09/03
Description.. Loads microsoft windows 24Bit file to memory.
Parameters... 1) Address of the device.
............. 2) File name.
............. 3) Pointer to BMP image info structre.
............. 4) Pointer to temp memory.
............. 5) Pointer to output memory.
Return....... Returns 0 when sucessfull.
............. Returns positive value if unsucessfull.
History...... 25/09/03 - Tidy up after review - MP
............. 17/09/03 - First Pass  - MP
-------------------------------------------------------------------*/
UBYTE	LoadMsWinBmp24(const DEVICE* dev, BYTE *filename,
					   BMPFILEINFO *bmpinfo, UWORD *memptrdatain,
					   ULONG dataoutaddress)
{
	BMPINFOHDR	bmpinfohdr;
	BMPFILEHDR	bmpfilehdr;
	BMPFILEINFO *bmpfileinfo;

	UBYTE	*membmpfiledata;

	UBYTE	reterror;

	/* Open File */
	if ( !(FAT32LoadFile(dev, filename, memptrdatain)) )
	{
		/* Get Byte Pointer to BMP data */
		(UBYTE *)membmpfiledata = (UBYTE *)memptrdatain;

		/* Get BMP file header */
		reterror = readbmpfilehdr(&bmpfilehdr, membmpfiledata);

		if ( reterror != 0 ) return(reterror);

		/* Get BMP info header */
		reterror = readbmpfileinfo(&bmpinfohdr, membmpfiledata);

		if ( reterror != 0 ) return(reterror);

		/* Get Pointer To BMP file information structure and populate */
		bmpfileinfo = bmpinfo;

		bmpfileinfo->imagewidth = bmpinfohdr.imagewidth;
		bmpfileinfo->imageheight = bmpinfohdr.imageheight;

		/* Get Data */
		readbmpdata24(&bmpfilehdr, &bmpinfohdr, membmpfiledata, dataoutaddress);

		return(0);
	}
	else
	{
		return(FILENOTFOUNDERR);
	}
}


/*-------------------------------------------------------------------
Name......... EnableAlphaBlendingLoading()
Author....... Aymeric
Date......... 08/09/05
Description.. Enable Bmp images to be loaded with Alpha Blending
............. characteristics (MSB set to 1 for each pixel)
Parameters... None
Return....... None.
-------------------------------------------------------------------*/

void EnableAlphaBlendingLoading()
{
	load_alphablended=TRUE;
}


/*-------------------------------------------------------------------
Name......... EnableAlphaBlendingLoading()
Author....... Aymeric
Date......... 08/09/05
Description.. Bmp images loaded normally (MSB set to 0 for each pixel)
Parameters... None
Return....... None.
-------------------------------------------------------------------*/
void DisableAlphaBlendingLoading()
{
	load_alphablended=FALSE;
}

/*-------------------------------------------------------------------
Name......... EnableAlphaBlendingLoading()
Author....... Aymeric
Date......... 08/09/05
Description.. return the status of loading
Parameters... None
Return....... TRUE => Alpha Blending enabled
............. FALSE => Bmp loaded normally
-------------------------------------------------------------------*/
UBYTE StatusAlphaBlendingLoading()
{
	return(load_alphablended);
}


/********************************************************************
*********************************************************************
						Static Functions
*********************************************************************
********************************************************************/

/*-------------------------------------------------------------------
Name......... readbmpfilehdr()
Author....... Mark Perkins
Date......... 17/09/03
Description.. Reads BMP file header information. BMP data is already
............. in memory
Parameters... 1) pointer to bmp header structure.
............. 2) pointer to BMP file data.
Return....... Returns 0 when sucessfull.
............. Returns positive value if unsucessfull.
History...... 25/09/03 - Tidy up afer review - MP
............. 17/09/03 - First Pass  - MP
-------------------------------------------------------------------*/
static UBYTE	readbmpfilehdr(BMPFILEHDR *bmphdr, UBYTE *bmpfiledata)
{
	BMPFILEHDR	*bmpfilehdr;

	register ULONG	dataoffset;

	/* Get pointer to bmp file header structure and populate */
	bmpfilehdr = bmphdr;

	bmpfilehdr->type[0] = *bmpfiledata++;
	bmpfilehdr->type[1] = *bmpfiledata++;

	/* Get pointer to data offset value */
	bmpfiledata += 8;

	dataoffset = (ULONG) *bmpfiledata++;		/* LSB*/

	dataoffset += ((ULONG) *bmpfiledata++) << 8;

	dataoffset += ((ULONG) *bmpfiledata++) << 16;

	dataoffset += ((ULONG) *bmpfiledata++) << 24;

	bmpfilehdr->dataoffset = dataoffset;

	/*
	Populate rest of structure with known values these elements of
	structure are never used by the driver
	*/
	bmpfilehdr->filesize = 0;
	bmpfilehdr->bfreserved[0] = 0;
	bmpfilehdr->bfreserved[1] = 0;
	bmpfilehdr->bfreserved[2] = 0;
	bmpfilehdr->bfreserved[3] = 0;

	if ( bmpfilehdr->type[0] == 'B' && bmpfilehdr->type[1] == 'M' )
	{
		return(0);
	}
	else
	{
		return(FILEFORMATERR);
	}
}

/*-------------------------------------------------------------------
Name......... readbmpfileinfo()
Author....... Mark Perkins
Date......... 17/09/03
Description.. Reads BMP file header information
Parameters... 1) Pointer to BMP info structure
............. 2) Pointer to bmp file data
Return....... Returns 0 when sucessfull.
............. Returns positive value if unsucessfull.
History...... 25/09/03 - Tidy up after review - MP
............. 17/09/03 - First Pass  - MP
-------------------------------------------------------------------*/
static UBYTE	readbmpfileinfo(BMPINFOHDR *bmpinfo, UBYTE *bmpfiledata)
{
	BMPINFOHDR *bmpinfohdr;

	register	ULONG	infosize;
	register	ULONG	imagewidth;
	register	ULONG	imageheight;
	register	ULONG	compression;

	ULONG	planes;
	ULONG	bitcount;

	/* Get data pointer to info data */
	bmpfiledata += BMPFILEHDRSIZE;

	/* Get pointer to bmp info header structure and populate */
	bmpinfohdr = bmpinfo;

	/* Get Info Block Size */
	infosize = *bmpfiledata++; 						/* LSB */
	infosize += ( (ULONG) *bmpfiledata++) << 8;
	infosize += ( (ULONG) *bmpfiledata++) << 16;
	infosize += ( (ULONG) *bmpfiledata++) << 24; 	/* MSB */

	bmpinfohdr->infosize = infosize;

	/* Get Image Width */
	imagewidth = *bmpfiledata++; 					/* LSB */
	imagewidth += ( (ULONG) *bmpfiledata++) << 8;
	imagewidth += ( (ULONG) *bmpfiledata++) << 16;
	imagewidth += ( (ULONG) *bmpfiledata++) << 24; 	/* MSB */

	bmpinfohdr->imagewidth = imagewidth;

	/* Get Image Height */
	imageheight = *bmpfiledata++; 					/* LSB */
	imageheight += ( (ULONG) *bmpfiledata++) << 8;
	imageheight += ( (ULONG) *bmpfiledata++) << 16;
	imageheight += ( (ULONG) *bmpfiledata++) << 24; /* MSB*/

	bmpinfohdr->imageheight = imageheight;

	/* Get Number of Planes */
	planes = *bmpfiledata++;
	planes += ( (ULONG) *bmpfiledata++) << 8;

	bmpinfohdr->planes = (UWORD) planes;

	/* Get Bit Count */
	bitcount = *bmpfiledata++;
	bitcount += ( (ULONG) *bmpfiledata++) << 8;

	bmpinfohdr->bitcount = bitcount;

	/* Get Compression */
	compression = *bmpfiledata++; 					/* LSB */
	compression += ( (ULONG) *bmpfiledata++) << 8;
	compression += ( (ULONG) *bmpfiledata++) << 16;
	compression += ( (ULONG) *bmpfiledata++) << 24; /* MSB */

	bmpinfohdr->compression = compression;

	/*
	Remaining Elements are unused by this driver so fill with
	known values
	*/
	bmpinfohdr->imagesize = 0;
	bmpinfohdr->xpixpermeter = 0;
	bmpinfohdr->ypixpermeter = 0;
	bmpinfohdr->colours = 0;
	bmpinfohdr->importantcolours = 0;


	if ( bmpinfohdr->bitcount != 24 )
	{
		return(FILEFORMATERR);
	}
	else if ( bmpinfohdr->planes != 1)
	{
		return(FILEFORMATERR);
	}
	else if ( bmpinfohdr->compression)
	{
		return(FILEFORMATERR);
	}
	else
	{
		return(0);
	}
}

/*-------------------------------------------------------------------
Name......... readbmpdata24()
Author....... Mark Perkins
Date......... 16/09/03
Description.. Reads BMP data and convert 24bit image to 16bit data
Parameters... 1) device
............. 2) file descriptor
............. 3) pointer to memory location
............. 4) pointer to BMP info structure
Return....... None.
History...... 25/09/03 - Tidy up after review - MP
............. 17/09/03 - First Pass  - MP
............. 04/01/05 - modified to fix colour 0 transparency problem AA
............. 08/09/05 - Modified to allow alpha blending loading AD
............. (MSB set to 0 or 1 according to load_alphablended flag)
-------------------------------------------------------------------*/
static void readbmpdata24(BMPFILEHDR *bmphdr, BMPINFOHDR *bmpinfo,
						  UBYTE *bmpfiledata, ULONG dataoutaddress)
{
	BMPINFOHDR *bmpinfohdr;
	BMPFILEHDR	*bmpfilehdr;

	register ULONG	colour;

	register ULONG	memrowaddress;
	register UWORD	*memrowfillptr;

	register ULONG	imagewidth;
	register ULONG	imageheight;

	register ULONG	doubleimagewidth;

	register ULONG	widthcount;

	register ULONG	red;
	register ULONG	green;
	register ULONG	blue;

	ULONG alphablendingmask;

	/* Get Pointer to file header structure */
	bmpfilehdr = bmphdr;

	/* Get BMP data pointer to RGB data */
	bmpfiledata	+= bmpfilehdr->dataoffset;

	/* Get pointer to info structure */
	bmpinfohdr = bmpinfo;

	/* Take local copy of image height and width */
	imagewidth = bmpinfohdr->imagewidth;

	imageheight = bmpinfohdr->imageheight;

	doubleimagewidth = imagewidth;
	doubleimagewidth += imagewidth;

	/*
	Calculate last row address as buffer needs to be filled backwards
	and bmp data is upside down
	*/
	(ULONG)memrowaddress = (ULONG)(dataoutaddress +
							((imageheight*imagewidth) * 2) ) - doubleimagewidth;

	if(load_alphablended)
	{
		alphablendingmask=ALPHABLENDING_MASK;
	}
	else
	{
		alphablendingmask=NONE_ALPHABLENDING_MASK;
	}

	/* code without alpha blending */
	while(imageheight)
	{
		/* Get Row Ptr */
		(ULONG)memrowfillptr = (ULONG)memrowaddress;

		/* Get Width */
		widthcount = imagewidth;

		while(widthcount)
		{
			/* Get Blue/Green/Red data */
			blue = (ULONG) *bmpfiledata++;
			green = (ULONG) *bmpfiledata++;
			red = (ULONG) *bmpfiledata++;

			/* colour zero transparency fix */
			if(!blue && !green && !red)	/* the pixel is REALLY black (r=g=b=0) */
				colour = 0;

			else {
				/* Convert 24bit data to 16bit */
				colour = (ULONG) (blue & 0xf8) >> 3;
				colour |= (ULONG) (green & 0xf8) << 2;
				colour |= (ULONG) (red & 0xf8) << 7;

				if (!colour)	/* pixel black after lower bit truncation */
					colour = VERY_DARK_GREY; /* pixel NOT REALLY black so set r=g=b=1 */
			}


			/*alphablend pixel or not */
			colour|=alphablendingmask;
			/* Fill buffer with byte swap for endian conversion */
			*memrowfillptr++ = (UWORD) BYTESWAPWORD( ((UWORD) colour) );

			widthcount--;
		}

		/* Remove Incerted Bytes */
		widthcount = imagewidth%4;

		while(widthcount)
		{
			*bmpfiledata++;
			widthcount--;
		}


		/* Decrement row ptr */
		memrowaddress -= doubleimagewidth;

		/* Decrement Height */
		imageheight--;
	}

}

