/*-------------------------------------------------------------------
Name......... tga.c
Author....... Mark Perkis
Date......... 11/09/03
Description.. TGA file loader.
-------------------------------------------------------------------*/

/*-------------------------------------------------------------------
DECLARATION FORMATS
Type defs........... UPPER CASE
#defines............ UPPER CASE
Global Functions.... GlobalFunction();
Static Functions.... staticfunction();
Global Variables.... GlobalVariable;
Static Variables.... staticvariable;
Local Variables..... localvariable;
Structure Elements.. Structure.element;
-------------------------------------------------------------------*/

/****	Include Files */
#include	"hardware.h"
#include	"interface.h"
#include	"tga.sp"

/****	Externals */

/****	Local Definitions */

/****	Global Variables */

/****	Local Variables */

/****	Global Constants */

/****	Local Constants */

/********************************************************************
*****	Functions
********************************************************************/
/*-------------------------------------------------------------------
Name......... LoadTGA()
Author....... Mark Perkins
Date......... 18/09/03
Description.. Loads a 16bpp TGA file to memory.
Parameters... 1) Address of the device
............. 2) Filename
............. 3) Pointer to TGA file info structure
............. 4) Pointer to temp memory areas.
............. 5) Pointer to output memory address.
Return....... Returns 0 if successfull
............. Returns code if failed
History...... 25/09/03 - Tidy up  - MP
............. 18/09/03 - First Pass  - MP
-------------------------------------------------------------------*/
UBYTE	LoadTGA16(const DEVICE* dev, BYTE *filename,
				  TGAFILEINFO *tgainfo, UWORD *memptrdatain,
				  ULONG dataoutaddress)
{

	TGAFILEHDR		tgafilehdr;
	TGAFILEFOOTER 	tgafilefooter;
	TGAFILEINFO 	*tgafileinfo;

	UBYTE	*memfiledata;

	UBYTE	reterror;

	/* Open File */
	if ( !(FAT32LoadFile(dev, filename, memptrdatain)))
	{
		/* Get Byte Pointer to BMP data */
		(UBYTE *)memfiledata = (UBYTE *)memptrdatain;

		/* Get TGA file header */
		reterror = readtgafilehdr(&tgafilehdr, memfiledata);

		if ( reterror != 0 )
		{
			return(reterror);
		}

		/* Get TGA file footer */
		reterror = readtgafilefooter(&tgafilefooter, &tgafilehdr,
									 memfiledata);

		if ( reterror != 0 )
		{
			return(reterror);
		}

		/* Get Pointer To TGA file information structure and populate */
		tgafileinfo = tgainfo;

		tgafileinfo->imagewidth = tgafilehdr.width;
		tgafileinfo->imageheight = tgafilehdr.height;

		/* Get Data */
		readtgadata16(tgafileinfo, memptrdatain, dataoutaddress);
	}
	else
	{
		return(FILENOTFOUNDERR);
	}

	return (0);

}

/*-------------------------------------------------------------------
Name......... readtgafilehdr()
Author....... Mark Perkins
Date......... 17/09/03
Description.. Reads TGA file header information. TGA file is already
............. in memory by this point.
Parameters... 1) pointer to TGA header structure.
............. 2) pointer to file data.
Return....... Returns 0 if successfull
............. Returns code if failed
History...... 25/09/03 - Tidy up  - MP
............. 18/09/03 - First Pass  - MP
-------------------------------------------------------------------*/
static UBYTE	readtgafilehdr(TGAFILEHDR *tgahdr, UBYTE *filedata)
{
	TGAFILEHDR	*tgafilehdr;

	register UWORD	xorigin;
	register UWORD	yorigin;

	register UWORD	width;
	register UWORD	height;

	/* Get pointer to TGA file header structure and populate */
	tgafilehdr = tgahdr;

	/* Get ID */
	tgafilehdr->id = *filedata++;

	/* Get Colour Map Type */
	tgafilehdr->colourmaptype = *filedata++;

	/* Get Image Type and check */
	tgafilehdr->imagetype = *filedata++;

	if (tgafilehdr->imagetype != TGAUNCOMPTRUECOLOUR)
	{
		return(FILEFORMATERR);
	}

	/* Get Colour Map Specification */
	tgafilehdr->colourmapspec[0] = *filedata++;
	tgafilehdr->colourmapspec[1] = *filedata++;
	tgafilehdr->colourmapspec[2] = *filedata++;
	tgafilehdr->colourmapspec[3] = *filedata++;
	tgafilehdr->colourmapspec[4] = *filedata++;

	/* Get X Origin */
	xorigin = *filedata++;
	xorigin += ( (UWORD) *filedata++) << 8;

	tgafilehdr->xorigin = xorigin;

	/* Get Y Origin */
	yorigin = *filedata++;
	yorigin += ( (UWORD) *filedata++) << 8;

	tgafilehdr->yorigin = yorigin;

	/* Get Width */
	width = *filedata++;
	width += ( (UWORD) *filedata++) << 8;

	tgafilehdr->width = width;

	/* Get Height */
	height = *filedata++;
	height += ( (UWORD) *filedata++) << 8;

	tgafilehdr->height = height;

	/* Get Pixel Depth and check */
	tgafilehdr->pixeldepth = *filedata++;

	if (tgafilehdr->pixeldepth != 16 )
	{
		return(FILEFORMATERR);
	}
	else
	{
		return(0);
	}

}

/*-------------------------------------------------------------------
Name......... readtgafilefooter()
Author....... Mark Perkins
Date......... 18/09/03
Description.. Reads TGA file footer.
Parameters... 1) Pointer to TGA file footer structure.
............. 2) Pointer to TGA file header structure.
............. 3) Pointer to file data.
Return....... Returns 0 if successfull
............. Returns code if failed
History...... 25/09/03 - Tidy up  - MP
............. 18/09/03 - First Pass  - MP
-------------------------------------------------------------------*/
static UBYTE readtgafilefooter(TGAFILEFOOTER *tgafooter,
							   TGAFILEHDR *tgahdr, UBYTE *filedata)
{
	TGAFILEFOOTER *tgafilefooter;
	TGAFILEHDR	*tgafilehdr;

	/* Get Pointer to TGA file header structure */
	tgafilehdr = tgahdr;

	/* Get file pointer to end of file */
	filedata += TGAFILEHDRSIZE;

	filedata += ( (tgafilehdr->height * tgafilehdr->width) * 2);

	/* Get Pointer to TGA footer structure and populate */
	tgafilefooter = tgafooter;

	/* Get Extension Area bits */
	tgafilefooter->extarea[0] = *filedata++;
	tgafilefooter->extarea[1] = *filedata++;
	tgafilefooter->extarea[2] = *filedata++;
	tgafilefooter->extarea[3] = *filedata++;

	/* Get Developer offset Info */
	tgafilefooter->devoffset[0] = *filedata++;
	tgafilefooter->devoffset[1] = *filedata++;
	tgafilefooter->devoffset[2] = *filedata++;
	tgafilefooter->devoffset[3] = *filedata++;

	/* Get Signature */
	tgafilefooter->signature[0] = *filedata++;
	tgafilefooter->signature[1] = *filedata++;
	tgafilefooter->signature[2] = *filedata++;
	tgafilefooter->signature[3] = *filedata++;
	tgafilefooter->signature[4] = *filedata++;
	tgafilefooter->signature[5] = *filedata++;
	tgafilefooter->signature[6] = *filedata++;
	tgafilefooter->signature[7] = *filedata++;
	tgafilefooter->signature[8] = *filedata++;
	tgafilefooter->signature[9] = *filedata++;
	tgafilefooter->signature[10] = *filedata++;
	tgafilefooter->signature[11] = *filedata++;
	tgafilefooter->signature[12] = *filedata++;
	tgafilefooter->signature[13] = *filedata++;
	tgafilefooter->signature[14] = *filedata++;
	tgafilefooter->signature[15] = *filedata++;

	/* Add String Terminator Char */
	tgafilefooter->signature[16] = '\0';

	/* Get Know Char */
	tgafilefooter->ch = *filedata++;

	/* Get Zero */
	tgafilefooter->zero = *filedata++;

	if (
		tgafilefooter->signature[0] == 'T' &&
		tgafilefooter->signature[1] == 'R' &&
		tgafilefooter->signature[2] == 'U' &&
		tgafilefooter->signature[3] == 'E' &&
		tgafilefooter->signature[4] == 'V' &&
		tgafilefooter->signature[5] == 'I' &&
		tgafilefooter->signature[6] == 'S' &&
		tgafilefooter->signature[7] == 'I' &&
		tgafilefooter->signature[8] == 'O' &&
		tgafilefooter->signature[9] == 'N' &&
		tgafilefooter->signature[10] == '-' &&
		tgafilefooter->signature[11] == 'X' &&
		tgafilefooter->signature[12] == 'F' &&
		tgafilefooter->signature[13] == 'I' &&
		tgafilefooter->signature[14] == 'L' &&
		tgafilefooter->signature[15] == 'E'

	)
	{
		return(0);
	}
	else
	{
		return(FILEFORMATERR);
	}
}

/*-------------------------------------------------------------------
Name......... readtgadata16()
Author....... Mark Perkins
Date......... 18/09/03
Description.. Reads 16bpp TGA data.
Parameters... 1) Pointer to TGA file info structure
............. 2) Pointer to TGA file data.
............. 3) Pointer to memory output address.
Return....... None
History...... 25/09/03 - Tidy up  - MP
............. 18/09/03 - First Pass  - MP
-------------------------------------------------------------------*/
static void	readtgadata16(TGAFILEINFO *tgainfo, UWORD *memfiledata,
						  ULONG dataoutaddress)
{
	TGAFILEINFO *tgafileinfo;

	register UWORD	*memfillptr;
	register UWORD	*memreadptr;
	register ULONG	imagewidth;
	register ULONG	imageheight;
	register ULONG	doubleimagewidth;
	register ULONG	memrowaddress;
	register ULONG	widthcount;


	/* Get Pointer to tga file info */
	tgafileinfo = tgainfo;

	/* Get local copy of width */
	imagewidth = tgafileinfo->imagewidth;

	doubleimagewidth = imagewidth;
	doubleimagewidth += imagewidth;

	/* Get local copy of height */
	imageheight = tgafileinfo->imageheight;

	/*
	Calculate last row address as buffer needs to be filled backwards
	and bmp data is upside down
	*/
	memrowaddress = (ULONG)(dataoutaddress +
							((imageheight*imagewidth) * 2) ) -
							(doubleimagewidth);

	/* Create Local Word Pointers and point to data */
	(ULONG)memreadptr = (ULONG)memfiledata;
	(ULONG)memreadptr += (TGAFILEHDRSIZE);

	while(imageheight)
	{
		/* Get Width */
		widthcount = imagewidth;

		/* Get Row Ptr */
		(ULONG)memfillptr = memrowaddress;

		while(widthcount)
		{
			*memfillptr++ = *memreadptr++;

			/* Decrement */
			widthcount--;
		}

		/* Decrement row ptr */
		memrowaddress -= doubleimagewidth;

		/* Decrement height */
		imageheight--;
	}
}
